/*
 ****************************************************************
 *								*
 *			pass1.c					*
 *								*
 *	Anlise sinttica - primeiro passo			*
 *								*
 *	Verso	1.0.0, de 25.03.86				*
 *		4.1.0, de 24.08.01				*
 *								*
 *	Rotinas:						*
 *		pass1     regular1     getsize			*
 *								*
 *	Mdulo: AS-80386/486/PENTIUM				*
 *		Tradutores					*
 *		Categoria D					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2001 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>

#include <stdio.h>
#include <a.out.h>

#include "../h/common.h"
#include "../h/tokens.h"
#include "../h/scan.h"
#include "../h/id_sym_tb.h"
#include "../h/expr.h"
#include "../h/opnd.h"
#include "../h/msg.h"
#include "../h/as.h"
#include "../h/op_code.h"
#include "../h/proto.h"

/*
 ****** Variveis globais ***************************************
 */
entry SYMTB	*labtb[LABSZ], **labp;	/* Alinhamento dos rtulos */

extern	int	if_false_level;	/* No. de Condicionais FALSE aninhadas */

/*
 ****************************************************************
 *	Anlise Sinttica - Primeiro Passo			*
 ****************************************************************
 */
void
pass1 (void)
{
	MAJOR		major;
	SYMTB		*sp;
	OPCODE		*op;
	char		*begin;
	EXP		exp;

	labp = labtb;

	while (getline () >= 0)
	{
		major = gettoken ();

		/*
		 *	Trata rtulos
		 */
		if (major == ID && latoken->l_major == CL)
		{
			if (if_false_level)
				continue;

			if (labp >= &labtb [LABSZ])
			{
				msg
				(	ERRO, LONGMARK,
					"Nmero excessivo de rtulos "
					"consecutivos"
				);
				continue;
			}

			sp = putsym (token->l_name, token->l_len, S_DEFINE);

			if   (sp->s.s_type == REG)
			{
				msg
				(	ERRO, LONGMARK,
					"Nome de um registro: "
					"no pode ser usado como rtulo"
				);
				continue;
			}
			elif (sp->s.s_type != UNDEF)
			{
				msg
				(	ERRO, LONGMARK,
					"Rtulo duplicado"
				);
				continue;
			}

			sp->s.s_value = *lc;
			sp->s.s_type = section;
			sp->s.s_flags |= S_LBL;
			symblk++;

			*labp++ = sp;

			gettoken ();
			gettoken ();
		}
		elif (major == LID && latoken->l_major == CL)
		{
			if (if_false_level)
				continue;

			if (labp >= &labtb [LABSZ])
			{
				msg
				(	ERRO, LONGMARK,
					"Nmero excessivo de rtulos "
					"consecutivos"
				);
				continue;
			}

			sp = putsym (token->l_name, token->l_len, S_DEFINE);

			if (sp->s.s_type != UNDEF)
			{
				msg
				(	ERRO, LONGMARK,
					"Smbolo duplicado: ser ignorado"
				);
				continue;
			}

			sp->s.s_value = *lc;
			sp->s.s_type = section;
			sp->s.s_flags |= S_LBL|S_LOCAL;

			*labp++ = sp;

			gettoken ();
			gettoken ();
		}
		elif (major == ID && latoken->l_major == EQ)
		{
			/*
			 *	Analisa atribuies
			 */
			if (if_false_level)
				continue;

			sp = putsym (token->l_name, token->l_len, S_DEFINE);
			begin = token->l_begin;

			gettoken ();	/* Obtem "==" */
			gettoken ();	/* Obtem o primeiro token da expresso */

			/*
			 *	Analisa a Expresso
			 */
			expression (&exp);

			/*
			 *	No permite redefinies
			 */
			if (sp->s.s_type != UNDEF && exp.ex_type != ERR)
			{
				if (exp.ex_type == REAL)
				{
					if
					(	sp->s.s_type == REAL &&
						sp->s_rval == exp.ex_rval
					)
					{
						msg
						(	ADVR, MYMARK,
							"Redefinio ignorada",
							NOSTR,
							begin, exp.ex_end
						);
					}
				}
				elif
				(	sp->s.s_value == exp.ex_value &&
					sp->s.s_type  == exp.ex_type
				)
				{
					msg
					(	ADVR, MYMARK,
						"Redefinio ignorada",
						NOSTR, begin, exp.ex_end
					);
				}
				else
				{
					msg
					(	ERRO, MYMARK,
						"Redefinio invlida",
						NOSTR, begin, exp.ex_end
					);
				}
				continue;
			}

			if (exp.ex_type == UNDEF)
			{
				msg
				(	ERRO, EXPMARK,
					"Expresso contm varivel(is) indefinida(s)",
					NOSTR, &exp
				);
				sp->s.s_type = ABS;
				sp->s.s_value = 0;
			}
			elif (exp.ex_type == REAL)
			{
				sp->s.s_type = REAL;
				sp->s_rval   = exp.ex_rval;
			}
			elif (exp.ex_type != ERR)
			{
				sp->s.s_type = exp.ex_type;
				sp->s.s_value = exp.ex_value;
			}
			continue;

		}	/* end atribuio */

		/*
		 *	Acabaram os rtulos
		 */
		if (token->l_major == EOL)
			continue;

		if (token->l_major != ID)
		{
			msg (ERRO, LONGMARK, "Esperava instruo ou pseudo");
			continue;
		}

		/*
		 *	Analisa a Instruo
		 */
		if ((op = op_hash (token->l_name)) == NOOPCODE)
		{
			msg
			(	ERRO, LONGMARK,
				"Instruo invlida"
			);
			continue;
		}

		if (if_false_level && (op->o_flags & NOSKIP) == 0)
			continue;

		opcodeno++;

		switch (op->o_type)
		{
		    /*
		     *	    Instrues Regulares
		     */
		    case I_REGULAR:
			*lc += regular1 (op);
			break;

		    /*
		     *	    Pseudo-Instrues
		     */
		    case I_PSEUDO:
			pseudo1 (op->o_group);
			break;

#if (0)	/*************************************/
		    /*
		     *	    Ponto Flutuante
		     */
		    case I_FLOAT:
			*lc += float1 (op);
			break;
#endif	/*************************************/

		    /*
		     *	    Tipo desconhecido
		     */
		    default:
			msg
			(	COMP, LONGMARK,
				"Instruo de tipo %d desconhecido",
				op->o_type
			);
			break;

		}	/* end  switch (type) */

		/*
		 *	Reinicializa a Tabela Auxiliar para
		 *	Alinhamento de Rotulos e a de Prefixos
		 */
		labp = labtb;

	}	/* end while (getline () >= 0) */

}	/* end pass1 */
