/*
 ****************************************************************
 *								*
 *			do_status.c				*
 *								*
 *	Diversas funes de mudana de estado			*
 *								*
 *	Verso	3.0.0, de 28.10.93				*
 *		3.0.5, de 12.02.97				*
 *								*
 *	Mdulo: DOSMP						*
 *		Utilitrios Sofisticados			*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2000 NCE/UFRJ - tecle "man licena"	*
 * 		Inspirado no PLDOS de Norival Ribeiro Figueira	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>

#include <stdio.h>
#include <string.h>

#include "../h/common.h"
#include "../h/dos.h"
#include "../h/proto.h"

/*
 ******	Prottipos de funes ***********************************
 */
void	do_iso_help (void);
void	do_bin_help (void);
void	do_ro_help (void);
void	do_rw_help (void);
void	do_lfn_help (void);
void	do_stat_help (void);

/*
 ****************************************************************
 *	Definies globais					*
 ****************************************************************
 */
entry int	bin_mode = 0;	/* Modo binrio ao invs ISO */
entry int	lfn_mode = 1;	/* == 0 => 8.3 * != 0 => 255 (W 95) */

/*
 ****************************************************************
 *	Altera a representao de dados para ISO		*
 ****************************************************************
 */
void
do_iso (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = cmd_getopt (argv, "H")) >= 0)
	{
		switch (opt)
		{
		    case 'H':			/* Help */
			do_iso_help ();
			return;

		    default:			/* Erro */
			fprintf (stderr, "\n");
			do_iso_help ();
			return;
		}
	}

	/*
	 *	Pequena consistncia
	 */
	argc -= cmd_optind;
	argv += cmd_optind;

	if (argc != 0)
		{ do_iso_help (); return; }

	bin_mode = 0;

}	/* end do_iso */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
do_iso_help (void)
{
	fprintf
	(	stderr,
		"%s - altera o modo da representao de dados para ISO (ASCII)\n"
		"\nSintaxe:\n"
		"\t%s\n",
		cmd_nm, cmd_nm
	);

}	/* end do_iso_help */

/*
 ****************************************************************
 *	Altera a representao de dados para BINRIO		*
 ****************************************************************
 */
void
do_bin (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = cmd_getopt (argv, "H")) >= 0)
	{
		switch (opt)
		{
		    case 'H':			/* Help */
			do_bin_help ();
			return;

		    default:			/* Erro */
			fprintf (stderr, "\n");
			do_bin_help ();
			return;
		}
	}

	/*
	 *	Pequena consistncia
	 */
	argc -= cmd_optind;
	argv += cmd_optind;

	if (argc != 0)
		{ do_bin_help (); return; }

	bin_mode = 1;

}	/* end do_bin */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
do_bin_help (void)
{
	fprintf
	(	stderr,
		"%s - altera o modo da representao de dados para BINRIO\n"
		"\nSintaxe:\n"
		"\t%s\n",
		cmd_nm, cmd_nm
	);

}	/* end do_bin_help */

/*
 ****************************************************************
 *	Permite apenas leituras no sistema de arquivos		*
 ****************************************************************
 */
void
do_ro (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = cmd_getopt (argv, "H")) >= 0)
	{
		switch (opt)
		{
		    case 'H':			/* Help */
			do_ro_help ();
			return;

		    default:			/* Erro */
			fprintf (stderr, "\n");
			do_ro_help ();
			return;
		}
	}

	/*
	 *	Pequena consistncia
	 */
	argc -= cmd_optind;
	argv += cmd_optind;

	if (argc != 0)
		{ do_ro_help (); return; }

	if (fs_status == FS_MOUNT_RW)
	{
	   /***	put_fat (); ***/

		fs_status = FS_MOUNT_RO;
	}

}	/* end do_ro */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
do_ro_help (void)
{
	fprintf
	(	stderr,
		"%s - permite apenas leituras no sistema de arquivos DOS\n"
		"\nSintaxe:\n"
		"\t%s\n",
		cmd_nm, cmd_nm
	);

}	/* end do_ro_help */

/*
 ****************************************************************
 *	Permite leituras e escritas no sistema de arquivos	*
 ****************************************************************
 */
void
do_rw (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = cmd_getopt (argv, "H")) >= 0)
	{
		switch (opt)
		{
		    case 'H':			/* Help */
			do_rw_help ();
			return;

		    default:			/* Erro */
			fprintf (stderr, "\n");
			do_rw_help ();
			return;
		}
	}

	/*
	 *	Pequena consistncia
	 */
	argc -= cmd_optind;
	argv += cmd_optind;

	if (argc != 0)
		{ do_rw_help (); return; }

	if (rflag)
	{
		printf
		(	"%s: No posso permitir escritas, "
			"pois foi dada a opo \"-r\"\n",
			cmd_nm
		);
		return;
	}

	if (dev_ro)
	{
		printf
		(	"%s: No posso permitir escritas, "
			"pois \"%s\" est com proteo de escrita\n",
			cmd_nm, dev_nm
		);
		return;
	}

	if (fs_status == FS_MOUNT_RO)
		fs_status = FS_MOUNT_RW;

}	/* end do_rw */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
do_rw_help (void)
{
	fprintf
	(	stderr,
		"%s - permite leituras e escritas no sistema de arquivos DOS\n"
		"\nSintaxe:\n"
		"\t%s\n",
		cmd_nm, cmd_nm
	);

}	/* end do_rw_help */

/*
 ****************************************************************
 *	Altera a criao de nomes de arquivos			*
 ****************************************************************
 */
void
do_lfn (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = cmd_getopt (argv, "H")) >= 0)
	{
		switch (opt)
		{
		    case 'H':			/* Help */
			do_lfn_help ();
			return;

		    default:			/* Erro */
			fprintf (stderr, "\n");
			do_lfn_help ();
			return;
		}
	}

	/*
	 *	Pequena consistncia
	 */
	argc -= cmd_optind;
	argv += cmd_optind;

	if (argc != 1)
		{ do_lfn_help (); return; }

	if   (streq (argv[0], "off") || streq (argv[0], "8.3"))
		lfn_mode = 0;
	elif (streq (argv[0], "on") || streq (argv[0], "255"))
		lfn_mode = 1;
	elif (streq (argv[0], "W95") || streq (argv[0], "w95"))
		lfn_mode = 1;
	else
		{ do_lfn_help (); return; }

}	/* end do_lfn */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
do_lfn_help (void)
{
	fprintf
	(	stderr,
		"%s - altera o modo de criao dos nomes DOS\n"
		"\nSintaxe:\n"
		"\t%s [off|8.3] [on|255|w95|W95]\n",
		cmd_nm, cmd_nm
	);

}	/* end do_lfn_help */

/*
 ****************************************************************
 *	Informa a representao de dados corrente		*
 ****************************************************************
 */
void
do_stat (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = cmd_getopt (argv, "H")) >= 0)
	{
		switch (opt)
		{
		    case 'H':			/* Help */
			do_stat_help ();
			return;

		    default:			/* Erro */
			fprintf (stderr, "\n");
			do_stat_help ();
			return;
		}
	}

	/*
	 *	Pequena consistncia
	 */
	argc -= cmd_optind;
	argv += cmd_optind;

	if (argc != 0)
		{ do_stat_help (); return; }

	/*
	 *	Escreve o estado
	 */
	switch (fs_status) 
	{
	    case FS_CLOSED:
		printf
		(	"No h dispositivo DOS disponvel\n"
		);
		break;

	    case FS_OPEN:
		printf
		(	"O sistema de arquivos \"%s\" NO est montado\n",
			dev_nm
		);
		break;

	    case FS_MOUNT_RO:
	    case FS_MOUNT_RW:
		printf
		(	"Sistema de arquivos \"%s\" montado %s\n",
			dev_nm, fs_status == FS_MOUNT_RO ?
			"somente para LEITURAS" : "para leituras e ESCRITAS"
		);

		printf
		(	"Opo \"-r\" %s",
			rflag  ? "DADA\n" : "NO dada, "
		);

		if (!rflag) printf
		(	"dispositivo \"%s\" %s proteo de escrita\n",
			dev_nm, dev_ro ?  "COM" : "SEM"
		);

#if (0)	/*******************************************************/
		printf
		(	"Tabela de endereos dos clusters (FAT) %s\n",
			dirty_fat ? "NO atualizada" : "ATUALIZADA"
		);
#endif	/*******************************************************/

		break;

	    default:
		error ("Estado invlido: %d", fs_status);
		break;

	}	/* end switch */

	/*
	 *	Representao de dados
	 */
	printf
	(	"Modo de representao de dados %s\n",
		bin_mode ? "BINRIO" : "ISO (ASCII)"
	);

	/*
	 *	Criao de nomes nos diretrios
	 */
	printf
	(	"Modo de criao dos nomes DOS: Estilo %s\n",
		lfn_mode ? "Windows 95" : "tradicional (8.3)"
	);

}	/* end do_stat */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
do_stat_help (void)
{
	fprintf
	(	stderr,
		"%s - informa o estado do sistema de arquivos e a "
			"repres. de dados corrente\n"
		"\nSintaxe:\n"
		"\t%s\n",
		cmd_nm, cmd_nm
	);

}	/* end do_stat_help */
