/*
 ****************************************************************
 *								*
 *			getopt.c				*
 *								*
 *	Fornece as opes de um comando				*
 *								*
 *	Verso	3.0.0, de 28.10.93				*
 *		3.0.0, de 09.06.95				*
 *								*
 *	Mdulo: DOSMP						*
 *		Utilitrios Sofisticados			*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2000 NCE/UFRJ - tecle "man licena"	*
 * 		Inspirado no PLDOS de Norival Ribeiro Figueira	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>

#include <stdio.h>

#include "../h/common.h"
#include "../h/proto.h"

/*
 ****************************************************************
 *	Variveis globais					*
 ****************************************************************
 */
entry const char	*cmd_optarg;   	/* Argumento de uma opo */
entry int		cmd_optind;	/* Indice do argumento apos as opes */
entry int		cmd_optinit;	/* Na prxima chamada, inicializa */

/*
 ****************************************************************
 *	Variveis locais estaticas				*
 ****************************************************************
 */
static const char *optp;		/* Ponteiro para a Opo corrente */
static const char **argvv;		/* Ponteiro para o Argumento corrente */

/*
 ****************************************************************
 *	Ponto de entrada da rotina				*
 ****************************************************************
 */
int
cmd_getopt (const char *argv[], const char *optstring)
{
	int		c;
	const char	*cp;

	/*
	 *	(Re-)Inicializao
	 */
	if (cmd_optinit)
	{
		argvv = argv; 		optp = "" - 1;
		cmd_optind = 1; 	cmd_optinit = 0;
	}

	cmd_optarg = NOSTR; optp++;

	/*
	 *	Verifica se este argumento tem mais opes 
	 */
	if (optp[0] == '\0')
	{
		argvv++; optp = argvv[0];

		if (optp == NOSTR || optp[0] != '-')
		{
			cmd_optind = argvv - argv;
			return (EOF);
		}

		if (optp[1] == '\0')
		{
			cmd_optind = argvv - argv;
			return (EOF);
		}

		if (optp[1] == '-')
		{
			cmd_optind = argvv - argv + 1;
			return (EOF);
		}

		optp++;

	}

	c = optp[0];

	/*
	 *	Procura a opo na lista
	 */
	for (cp = optstring; cp[0] != '\0'; cp++) 
	{
		if (cp[0] == c)
			goto found;
	}

	fprintf (stderr, "%s: Opo invlida: '%c'\n", argv[0], c);

	cmd_optarg = optp; return ('?');

	/*
	 *	A opo foi encontrada na lista
	 */
    found:
	if (cp[1] != ':')
		return (c);

	/*
	 *	A Opo tem Argumento
	 */
	argvv++;

	if (optp[1] == '\0' && optp[-1] == '-')
	{
		/*
		 *	Retorna o argumento
		 */
		if ((cmd_optarg = argvv[0]) == NOSTR)
		{
			fprintf
			(	stderr,
				"%s: Faltando o argumento da opo '%c'\n",
				argv[0],
				c
			);

			cmd_optarg = optp; return ('?');
		}

		return (c);
	}
	else
	{
		/*
		 *	Foram encontrados Caracteres Estranhos
		 *	antes ou depois desta opo
		 */
		fprintf
		(	stderr,
			"%s: Caracteres estranhos antes ou "
			"aps a opo '%c'\n",
			argv[0],
			c
		);

		cmd_optarg = optp; return ('?');
	}

}	/* end cmd_getopt */
