/*
 ****************************************************************
 *								*
 *			emptydir.c				*
 *								*
 * 	Gera os nomes dos diretrios vazios (para remoo)	*
 *								*
 *	Verso	3.1.0, de 14.01.99				*
 *		4.2.0, de 08.05.02				*
 *								*
 *	Mdulo: emptydir					*
 *		Utilitrios experimentais			*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2002 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <ftw.h>
#include <fcntl.h>
#include <errno.h>

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
const char	pgversion[] =  "Verso:	4.2.0, de 08.05.02";

#define	EVER	;;
#define	elif	else if
#define STR(x)	# x
#define XSTR(x)	STR (x)

/*
 ******	Variveis globais ***************************************
 */
entry int	gflag;		/* Gera o garfield */
entry int	vflag;		/* Verbose */
entry int	Nflag;		/* Nomes dos pgms no <stdin> */

entry int	n_dirs;		/* No. total de diretrios vazios */

/*
 ****** Garfield ************************************************
 */
const char	garfield[] = "garfield";
entry FILE	*gfp;

/*
 ******	Prottipos de funes ***********************************
 */
int		accept (const char *, STAT *, int);
void		dir_read (const char *, STAT *);
void		help (void);

/*
 ****************************************************************
 * 	Gera os nomes dos diretrios vazios (para remoo)	*
 ****************************************************************
 */
int
main (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = getopt (argc, argv, "gvNH")) != EOF)
	{
		switch (opt)
		{
		    case 'g':			/* garfield */
			gflag++;
			continue;

		    case 'v':			/* verbose flag */
			vflag++;
			continue;

		    case 'N':			/* Nomes do <stdin> */
			Nflag++;
			break;

		    case 'H':			/* Help */
			help ();

		    default:			/* Erro */
			fputc ('\n', stderr);
			help ();

		}	/* end switch */

	}	/* end while */ 

	/*
	 *	Ajustes
	 */
   /***	argc -= optind; ***/
	argv += optind;

	/*
	 *	Prepara o "garfield"
	 */
	if (gflag && (gfp = fopen (garfield, "w")) == NOFILE)
		error ("$*No consegui abrir \"%s\"", garfield);

	/*
	 *	Analisa o Nflag
	 */
	if (Nflag && *argv != NOSTR)
	{
		error ("Os argumentos suprfluos sero ignorados");
		*argv = NOSTR;
	}

	/*
	 *	Procura os argumentos
	 */
	if (*argv == NOSTR) 		 /* No foram dados argumentos: l do "stdin" */
	{
		char		*area = alloca (1024);

		while (fngets (area, 1024, stdin) != NOSTR)
		{
			ftw (area, accept);
		}
	}
	else 				 /* Foram dados argumentos */
	{
		for (/* vazio */; *argv; argv++)
		{
			ftw (*argv, accept);
		}
	}

	/*
	 *	Imprime o total
	 */
	if (!vflag)
		return (0);

	if (n_dirs == 0)
		error ("No foram encontrados diretrios vazios");
	else
		error ("Foram encontrados %d diretrios vazios", n_dirs);

	return (0);

}	/* end emptydir */

/*
 ****************************************************************
 *	Processa um nome 					*
 ****************************************************************
 */
int
accept (const char *file_nm, STAT *sp, int status)
{
	if   (status == FTW_NS)
	{
		error ("*No consegui obter o estado de \"%s\"", file_nm);
		return (0);
	}
	elif (status == FTW_DNR)
	{
		error ("*No consegui percorrer \"%s\"", file_nm);
	}

	/*
	 *	S interessam diretrios
	 */
	if ((sp->st_mode & S_IFMT) == S_IFDIR)
		dir_read (file_nm, sp);

	return (0);

}	/* end accept */

/*
 ****************************************************************
 *	Le um diretrio						*
 ****************************************************************
 */
void
dir_read (const char *dir_nm, STAT *sp)
{
	DIR		*dir_fp;
	const DIRENT	*dp;

	/*
	 *	Abre o diretrio
	 */
	if ((dir_fp = opendir (dir_nm)) == NODIR)
	{
		error ("*No consegui abrir \"%s\"", dir_nm);
		return;
	}

	/*
	 *	Le o diretrio
	 */
	while ((dp = readdir (dir_fp)) != NODIRENT)
	{
		if (streq (dp->d_name, ".") || streq (dp->d_name, ".."))
			continue;

		/* No est vazio */

		closedir (dir_fp); return;

	}	/* end while (fread) */

	/*
	 *	Est vazio
	 */
	closedir (dir_fp);

	n_dirs++;

	if (gflag)
		fprintf (gfp, "%s\n", dir_nm);
	else
		printf ("%s\n", dir_nm);

}	/* end dir_read */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
		"%s - gera os nomes dos diretrios vazios\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"\t%s [-gvN] [<rvore> ...]\n\n",
		pgname, pgversion, pgname
	);
	fprintf
	(	stderr,
		"\nOpes:"
		"\t-g: Gera o arquivo \"garfield\"\n"
		"\t-v: Verboso\n"
		"\t-N: L os nomes das <rvore>s de \"stdin\"\n"
	);
	exit (2);

}	/* end help */
