/*
 ****************************************************************
 *								*
 *			port_daemon.c				*
 *								*
 *	Port Mapper daemon					*
 *								*
 *	Verso	4.8.0, de 02.08.05				*
 *		4.8.0, de 02.08.05				*
 *								*
 *	Mdulo: ITNET						*
 *		Utilitrios de comunicao			*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2005 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/common.h>
#include <sys/sync.h>
#include <sys/itnet.h>
#include <sys/times.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <xti.h>
#include <errno.h>

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
const char	pgversion[] =  "Verso:	4.8.0, de 02.08.05";

#define	PMAP_PORT	111

/*
 ******	Converte um longo para BIG_ENDIAN (ou vice-versa) *******
 */
#define ENDIAN_CV(x)	(((x) << 24)		  |	\
			(((x) & 0x0000FF00) << 8) |	\
			(((x) & 0x00FF0000) >> 8) |	\
			(((unsigned)(x) >> 24)))

/*
 ******	Definies do protocoloes *******************************
 */
enum {	RPC_VERS = 2 };

enum {	RPC_CALL, RPC_REPLY };

enum {	RPC_MSG_ACCEPTED, RPC_MSG_DENIED };

enum {	IPPROTO_TCP = 6, IPPROTO_UDP = 17 };

enum {	PMAP_PROG = 100000, PMAP_VERS = 2 };

enum {	PMAPPROC_NULL, PMAPPROC_SET, PMAPPROC_UNSET, PMAPPROC_GETPORT, PMAPPROC_DUMP };

enum {	MOUNTPROG = 100005, MOUNTVERS = 1 };

enum {	MOUNTPROG_NULL, MOUNTPROG_MNT, MOUNTPROG_DUMP, MOUNTPROG_UMNT, MOUNTPROG_UMNTALL };




/*
 ******	Estrutura para o envio de um pedido de PORTA ************
 */
typedef struct
{
	int	r_xid;		/* No. do pedido */
	int	r_mtype;	/* Pedido/resposta */
	int	r_rpcver;	/* Verso do protocolo do RPC */
	int	r_prog;		/* No. do programa do PORT MAPPER */
	int	r_vers;		/* Verso do protocolo do PORT MAPPER */
	int	r_proc;		/* Funo pedida */

	int	r_cred[2];	/* Autenticao */
	int	r_verf[2];

	int	r_req_prog;	/* Dados do programa do qual desejamos  */
	int	r_req_vers;	/* saber a porta */
	int	r_req_prot;
	int	r_req_port;

}	PMAP_CALL;

/*
 ******	Estrutura para o recebimento de uma PORTA ***************
 */
typedef struct
{
	int	r_xid;		/* No. do pedido */
	int	r_mtype;	/* Pedido/resposta */
	int	r_rstat;	/* A Mensagem foi aceita? */

	int	r_verf[2];

	int	r_astat;	/* O RPC foi executado? */
	int	r_port;		/* Porta pedida */

}	PMAP_REPLY;







#if (0)	/*******************************************************/
#define	TIMEOUT		10	/* Segundos */
#define	MILHAO		1000000

#define	EVER	;;
#define	elif	else if
#define STR(x)	# x
#define XSTR(x)	STR (x)

typedef unsigned int	bit;
#endif	/*******************************************************/

#if (0)	/*******************************************************/
/*
 ******	Verifica se as macro variveis esto definidas **********
 */
#if	!defined (BIG_ENDIAN) && !defined (LITTLE_ENDIAN)
#error	"Uma das macrovariveis \"BIG/LITTLE_ENDIAN\" deve estar definida"
#endif	BIG_ENDIAN/LITTLE_ENDIAN

/*
 ******	Definies relativas  ordem de bytes da CPU ************
 */
#ifdef	LITTLE_ENDIAN
#define	ENDIAN_LONG(x)	long_endian_cv (x)
#define	ENDIAN_SHORT(x)	short_endian_cv (x)
#else
#define	ENDIAN_LONG(x)	(x)
#define	ENDIAN_SHORT(x)	(x)
#endif	LITTLE_ENDIAN
#endif	/*******************************************************/

/*
 ****** Variveis globais ***************************************
 */
entry int		vflag = 0;	/* Verbose */

const int		tst = ENDIAN_CV (100000);

entry char		error_msg_to_log; /* No nulo se mensagens para "itnetlog" */

const char		udp_dev[] = "/dev/itnudp";

entry int		udp_fd;		/* fd do UDP */

entry long		pidfilho;	/* Para enviar o sinal */

#if (0)	/*******************************************************/
entry const char	*area_ptr;	/* para "get_domain_name" */
#endif	/*******************************************************/

#if (0)	/*******************************************************/
/*
 ****** Tabela de servidores de nomes ***************************
 */
#define N_SERVTB	10	/* No. mximo de servidores */	

typedef enum
{ 
	/* Repare os cdigos dos servidores crescentes em qualidade */

	SERVER_NO_SERVER_AT_ALL = 0,	/* No tenho nenhum servidor disponvel */
	SERVER_NO_SERVER_ON_ADDR,	/* Neste endereo no tem servidor */
	SERVER_TIMEOUT,			/* No respondeu */
	SERVER_BAD_ANSWER,		/* O servidor deu uma resposta intil */
	SERVER_OK,			/* Obteve o endereo desejado */
	SERVER_NULL 			/* No usado ainda */

}	SERVER_ENUM;

typedef struct
{
	IPADDR	s_addr;		/* Endereo do Servidor */

	char	s_index;	/* No. do servidor */
	char	s_code;		/* Cdigo de servio: veja acima */
	bit	s_checksum : 1;	/* A resposta tem "checksum" */

	long	s_time;		/* Tempo de resposta em micro-segundos */
	MUTM	s_mutm;		/* Data de envio da consulta */

}	SERVTB;

entry SERVTB	servtb[N_SERVTB+1];	/* A tabela de servidores */	

/*
 ****** Header Section Format ***********************************
 */
typedef struct
{
	ushort	h_id;		/* Identificao */

#ifdef	LITTLE_ENDIAN
	bit	h_rd : 1;	/* Recursion desired */
	bit	h_tc : 1;	/* Truncation */
	bit	h_aa : 1;	/* Authorative Answer */
	bit	h_op : 4;	/* Kind of Query */
	bit	h_qr : 1;	/* Query or response */

	bit	h_rc : 4;	/* Response code */
	bit	h_z  : 3;	/* (reserved) */
	bit	h_ra : 1;	/* Recursion available */
#else
	bit	h_qr : 1;	/* Query or response */
	bit	h_op : 4;	/* Kind of Query */
	bit	h_aa : 1;	/* Authorative Answer */
	bit	h_tc : 1;	/* Truncation */
	bit	h_rd : 1;	/* Recursion desired */

	bit	h_ra : 1;	/* Recursion available */
	bit	h_z  : 3;	/* (reserved) */
	bit	h_rc : 4;	/* Response code */
#endif	LITTLE_ENDIAN

	ushort	h_qdcnt;	/* No. de entradas */
	ushort	h_ancnt;	/* No. de RRs de resposta */
	ushort	h_nscnt;	/* No. de RRs de autoridade */
	ushort	h_arcnt;	/* No. de RRs adicionais */

}	H;

/*
 ****** Tabela com os tipos de erros do DOMAIN ******************
 */
enum DOMAIN_ERROR_CODE
{
	/*  0 */   DOMAIN_NO_ERROR = 0,		/* Nenhum erro */
	/*  1 */   DOMAIN_FORMAT_ERROR,		/* O servidor no entendeu */
	/*  2 */   DOMAIN_SERVER_FAILURE,	/* Problemas com o servidor */
	/*  3 */   DOMAIN_NAME_ERROR,		/* Nome no existe */
	/*  4 */   DOMAIN_NOT_IMPLEMENTED,	/* Servio inexistente */
	/*  5 */   DOMAIN_REFUSED		/* O servidor no faz este servio */

	/*  6 a 15 Reservado para uso futuro */
};

/*
 ****** Tabela com os nomes dos tipos do DOMAIN *****************
 */
enum TYPE_CODE
{
	/*  1 */   TYPE_A = 1,	/* Host address */
	/*  2 */   TYPE_NS,	/* Authoritative name server */
	/*  3 */   TYPE_ND,	/* Mail destination (obsolete) */
	/*  4 */   TYPE_MF,	/* Mail forwarder (obsolete) */
	/*  5 */   TYPE_CNAME,	/* Canonical name for an alias */
	/*  6 */   TYPE_SOA,	/* Start of a zone of authority */
	/*  7 */   TYPE_MB,	/* Mailbox domain name (experimental) */
	/*  8 */   TYPE_MG,	/* Mail group member (experimental) */
	/*  9 */   TYPE_MR,	/* Mail rename domain name (experimental) */
	/* 10 */   TYPE_NULL,	/* NULL RR (experimental) */
	/* 11 */   TYPE_WKS,	/* Well known service description */
	/* 12 */   TYPE_PTR,	/* Domain name pointer */
	/* 13 */   TYPE_HINFO,	/* Host information */
	/* 14 */   TYPE_MINFO,	/* Mailbox or mail list information */
	/* 15 */   TYPE_MX,	/* Mail exchange */
	/* 16 */   TYPE_TXT	/* Text strings */
};

typedef struct
{
	char	t_code[8];	/* Cdigo do tipo */
	char	*t_name;	/* Descrio extensa do tipo */

}	TYPE_DESC;

const TYPE_DESC	type_desc[] =
{
	/*  0 */   "",		"Type ZERO",
	/*  1 */   "A",		"Host address",
	/*  2 */   "NS",	"Authoritative name server",
	/*  3 */   "ND",	"Mail destination (obsolete)",
	/*  4 */   "MF",	"Mail forwarder (obsolete)",
	/*  5 */   "CNAME",	"Canonical name for an alias",
	/*  6 */   "SOA",	"Start of a zone of authority",
	/*  7 */   "MB",	"Mailbox domain name (experimental)",
	/*  8 */   "MG",	"Mail group member (experimental)",
	/*  9 */   "MR",	"Mail rename domain name (experimental)",
	/* 10 */   "NULL",	"NULL RR (experimental)",
	/* 11 */   "WKS",	"Well known service description",
	/* 12 */   "PTR",	"Domain name pointer",
	/* 13 */   "HINFO",	"Host information",
	/* 14 */   "MINFO",	"Mailbox or mail list information",
	/* 15 */   "MX",	"Mail exchange",
	/* 16 */   "TXT",	"Text strings",
	/*    */   "",		NOSTR
};

/*
 ****** Tabela com os nomes das classes do DOMAIN ***************
 */
enum CLASS_CODE
{
	/*  1 */   CLASS_IN = 1,/* Internet */
	/*  2 */   CLASS_CS,	/* CSNET class (obsolete) */
	/*  3 */   CLASS_CH,	/* CHAOS class */
	/*  4 */   CLASS_HS	/* HESIOD class */
};

typedef struct
{
	char	c_code[8];	/* Cdigo da classe */
	char	*c_name;	/* Descrio extensa da classe */

}	CLASS_DESC;

const CLASS_DESC class_desc[] =
{
	/*  0 */   "",		"Class ZERO",
	/*  1 */   "IN",	"Internet",
	/*  2 */   "CS",	"CSNET class (obsolete)",
	/*  3 */   "CH",	"CHAOS class",
	/*  4 */   "HS",	"HESIOD class",
	/*    */   "",		NOSTR
};
#endif	/*******************************************************/

/*
 ****** Prottipos de funes ***********************************
 */


int		send_req (void);
int		receive_answer (void);
void		error (const char *fmt, ...);
void		help (void);







#if (0)	/*******************************************************/
void		filho (void);
void		pai (void);
void		analyse_udp_datagrams (SERVTB *, char *);
char		*get_domain_name (const char **cpp);
void		put_domain_name (char *name, char **area);
int		get_short (const char **cpp);
void		put_short (int value, char **cpp);
long		get_long (const char **cpp);
IPADDR		t_arpa_str_to_addr (const char *);
const char	*get_type_name (int);
const char	*get_class_name (int);
char		*h_op_text (int);
char		*h_rc_text (enum DOMAIN_ERROR_CODE);
char		*edit_kernel_code (enum DNS_ERROR_CODE);
void		on_signal (int, ...);
#endif	/*******************************************************/

/*
 ****************************************************************
 *	Domain name resolver					*
 ****************************************************************
 */
void
main (int argc, const char *argv[])
{
	int			opt;
#if (0)	/*******************************************************/
	SERVTB			*sp;
	int			index;
#endif	/*******************************************************/
	INADDR			server_addr;
	T_BIND			bind;

	/*
	 *	Pequena inicializao
	 */
	error_msg_to_log++;

	/*
	 *	Verifica se  SUPERUSURIO
	 */
	if (geteuid () != 0)
		error ("$O usurio efetivo no  SUPERUSURIO");

	/*
	 *	Analisa as opes
	 *
	 *	Sintaxe:
	 *		port_daemon [-v] <port_server_addr> ...
	 */
	while ((opt = getopt (argc, argv, "v")) != EOF)
	{
		switch (opt)
		{
		    case 'v':			/* Verbose */
			vflag++;
			break;

		    default:			/* Erro */
			putc ('\n', stderr);
			help ();

		}	/* end switch */

	}	/* end while */

	argv += optind;
	argc -= optind;

#if (0)	/*******************************************************/
	if (argc == 0)
		help ();
#endif	/*******************************************************/

	/*
	 *	Abre os "endpoint"s
	 */
	if ((udp_fd = t_open (udp_dev, O_RDWR, (T_INFO *)NULL)) < 0)
		error ("$*No consegui abrir \"%s\"", udp_dev);

	/*
	 *	Obtm a porta local
	 */
	server_addr.a_port = PMAP_PORT;
	server_addr.a_addr = 0;

	bind.addr.len	 = sizeof (INADDR);
	bind.addr.maxlen = sizeof (INADDR);
	bind.addr.buf	 = &server_addr;

	if (t_bind (udp_fd, &bind, &bind) < 0)
		error ("$*No consegui dar \"t_bind\" UDP");

	send_req ();

	receive_answer ();

#if (0)	/*******************************************************/
	/*
	 *	Cria um filho para enviar os pedidos de DNS.
	 *	O pai le os datagramas de resposta.
	 */
	if ((pidfilho = thread ()) < 0)
		error ("$*No consegui criar um novo processo");

	if (pidfilho > 0)
		pai ();
	else
		filho ();

	/* No retorna */
#endif	/*******************************************************/

}	/* end port_mapper */

/*
 ****************************************************************
 *	Envia o pedido						*
 ****************************************************************
 */
int
send_req (void)
{
	T_UNITDATA	ud;
	INADDR		client_addr;
	IPADDR		dst_addr;
	const char	*dst_nome_nm = "presto";
	static int	id;
	PMAP_CALL	dg;

	/*
	 *	Prepara o datagrama
	 */
	memset (&dg, 0, sizeof (dg));

	dg.r_xid	= long_endian_cv (++id);
	dg.r_mtype	= ENDIAN_CV (RPC_CALL);
	dg.r_rpcver	= ENDIAN_CV (RPC_VERS);
	dg.r_prog	= ENDIAN_CV (PMAP_PROG);
	dg.r_vers	= ENDIAN_CV (PMAP_VERS);
	dg.r_proc	= ENDIAN_CV (PMAPPROC_GETPORT);

   /***	dg.r_cred	= 0;	***/
   /***	dg.r_verf	= 0;	***/

	dg.r_req_prog	= ENDIAN_CV (MOUNTPROG);
	dg.r_req_vers	= ENDIAN_CV (MOUNTVERS);
	dg.r_req_prot	= ENDIAN_CV (IPPROTO_UDP);
	dg.r_req_port	= 0;	/* Ignorado */

	/*
	 *	Obtm o endereo do n remoto
	 */
	if ((dst_addr = t_node_to_addr (udp_fd, dst_nome_nm, (long *)NULL)) == -1)
		error ("$*No consegui obter o endereo de \"%s\"", dst_nome_nm);

	/*
	 *	Envia o datagrama UDP
	 */
	client_addr.a_addr = dst_addr;
	client_addr.a_port = PMAP_PORT;

	ud.addr.len	= sizeof (INADDR);
   	ud.addr.buf	= &client_addr;

	ud.udata.len	= sizeof (dg);
	ud.udata.buf	= &dg;

	ud.opt.len	= 0;

	if (t_sndudata (udp_fd, &ud) < 0)
		error ("$*No consegui enviar o datagrama");

#if (0)	/*******************************************************/
	{
		int fd = creat ("dg", 0644);

		write (fd, &dg, sizeof dg);

	}
#endif	/*******************************************************/

	return (0);

}	/* end send_req */

/*
 ****************************************************************
 *	Recebe a resposta					*
 ****************************************************************
 */
int
receive_answer (void)
{
	T_UNITDATA	ud;
	INADDR		client_addr;
	int		flags;
	PMAP_REPLY	dg;




#if (0)	/*******************************************************/
typedef struct
{
	int	r_xid;		/* No. do pedido */
	int	r_mtype;	/* Pedido/resposta */
	int	r_rstat;	/* A Mensagem foi aceita? */

	int	r_verf[2];

	int	r_astat;	/* O RPC foi executado? */
	int	r_port;		/* Porta pedida */

}	PMAP_REPLY;
#endif	/*******************************************************/


	/*
	 *	x
	 */
	ud.addr.maxlen	= sizeof (INADDR);
   	ud.addr.buf	= &client_addr;

	ud.udata.maxlen	= sizeof (dg);
	ud.udata.buf	= &dg;

	ud.opt.len	= 0;

	/*
	 *	x
	 */
	if (t_rcvudata (udp_fd, &ud, &flags) < 0)
		error ("$*No consegui receber o datagrama");

printf ("receive_answer: Recebi %d bytes\n", ud.udata.len);

	{
		int fd = creat ("rcv.dg", 0644);

		write (fd, &dg, ud.udata.len);

	}

	/*
	 *	Analisa a resposta
	 */
printf ("receive_answer: Porta = %d\n", long_endian_cv (dg.r_port));









	return (0);

}	/* end receive_answer */










#if (0)	/*******************************************************/
/*
 ****************************************************************
 *	Filho: Espera receber novos pedidos			*
 ****************************************************************
 */
void
filho (void)
{
	SERVTB		*sp;
	DNS		dns;

	for (EVER)
	{
		if (ioctl (udp_fd, I_DNS_WAIT_REQ, &dns) < 0)
			error ("*Erro no \"ioctl\" I_DNS_WAIT_REQ");

		if (vflag) error
		(	"Recebendo pedido do endereo de \"%s\"",
			dns.d_host_nm
		);

		for (sp = servtb; sp->s_addr != 0; sp++)
		{
			if (vflag) error
			(	"Consultando a \"%s\"",
				t_addr_to_str (sp->s_addr)
			);

			send_req (sp, &dns);
		}

	}	/* end for (EVER) */

}	/* end filho */
#endif	/*******************************************************/

#if (0)	/*******************************************************/
/*
 ****************************************************************
 *	Envia o pedido						*
 ****************************************************************
 */
int
send_req (SERVTB *sp, DNS *np)
{
	T_UNITDATA			unitdata;
	INADDR				addr;
	struct { H h; char c[256]; }	dg;
	char				*str;
	int				sz;
	char				area[128];
	static int			id;

	/*
	 *	Prepara o datagrama
	 */
	memset (&dg, 0, sizeof (dg));

	dg.h.h_id = ENDIAN_SHORT (++id);
   /***	dg.h.h_qr = 0;	***/
   /***	dg.h.h_op = 0;	***/
   /***	dg.h.h_aa = 0;	***/
   /***	dg.h.h_tc = 0;	***/
	dg.h.h_rd = 1;
   /***	dg.h.h_ra = 0;	***/
   /***	dg.h.h_z  = 0;	***/
   /***	dg.h.h_rc = 0;	***/
	dg.h.h_qdcnt = ENDIAN_SHORT (1);
   /***	dg.h.h_ancnt = 0;	***/
   /***	dg.h.h_nscnt = 0;	***/
   /***	dg.h.h_arcnt = 0;	***/

	str = dg.c;

	/*
	 *	Compe o nome reverso, se for o caso
	 */
	if (np->d_entry_code == 'R')
	{
		const char	*cp = (char *)&np->d_host_addr;
#ifdef	LITTLE_ENDIAN
		sprintf (area, "%u.%u.%u.%u.IN-ADDR.ARPA", cp[0], cp[1], cp[2], cp[3]);
#else
		sprintf (area, "%u.%u.%u.%u.IN-ADDR.ARPA", cp[3], cp[2], cp[1], cp[0]);
#endif	LITTLE_ENDIAN

		put_domain_name (area, &str);
	}
	else
	{
		put_domain_name (np->d_host_nm, &str);
	}

	if   (np->d_entry_code == 'R')
		put_short (TYPE_PTR, &str);	/* Type = PTR */
	elif (np->d_entry_code == 'M')
		put_short (TYPE_MX, &str);	/* Type = MX */
	elif (np->d_entry_code == 'C')
		put_short (TYPE_A, &str);	/* Type = A */
	else
		return (0);

	put_short (CLASS_IN, &str);	/* Class = IN */

	sz = str - (char *)&dg;

	/*
	 *	Manda o datagrama UDP
	 */
	addr.a_addr = sp->s_addr;
	addr.a_port = DOMAIN_PORT;

	unitdata.addr.len	= sizeof (addr);
   	unitdata.addr.buf	= &addr;

	unitdata.udata.len	= sz;
	unitdata.udata.buf	= &dg;

	mutime (&sp->s_mutm);

	if (t_sndudata (udp_fd, &unitdata) >= 0)
	{
		if (vflag) error
		(
			"Enviando UDP (id %d) para %s", 
			id, t_addr_to_str (sp->s_addr)
		);

		return (0);
	}

	/*
	 *	No conseguiu enviar
	 */
   /***	strcpy (np->d_host_nm, ...); ***/
   	np->d_host_addr = 0;

   /***	np->d_entry_code = ...; ***/
   /***	np->d_resolver_done = ...; ***/
	np->d_error_code = DNS_SERVER_ERROR;
	np->d_server_index = sp->s_index;

	np->d_expir_time = 60;
   /***	np->d_last_use_time = ...; ***/

	sp->s_code = SERVER_NO_SERVER_ON_ADDR;

	if (ioctl (udp_fd, I_DNS_PUT_INFO, np) < 0)
		error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");

	error ("*Erro no SNDUDP");

	return (-1);

}	/* end send_req */
#endif	/*******************************************************/

#if (0)	/*******************************************************/
/*
 ****************************************************************
 *	Espera uma resposta					*
 ****************************************************************
 */
void
pai (void)
{
	SERVTB		*sp;
	T_UNITDATA	unitdata;
	char 		area[512 + 128];
	int		flag;
	INADDR		rcv_addr;
	MUTM		end_time;

	/*
	 *	Para terminar elegantemente
	 */
	signal (SIGINT,  on_signal);
	signal (SIGQUIT, on_signal);
	signal (SIGTERM, on_signal);
	signal (SIGHUP,  on_signal);

	/*
	 *	Espera datagramas UDP
	 */
	for (EVER)
	{
	    again:
		unitdata.addr.maxlen	= sizeof (rcv_addr);
		unitdata.addr.buf	= &rcv_addr;

		unitdata.udata.maxlen	= sizeof (area);
		unitdata.udata.buf	= area;

		if (t_rcvudata (udp_fd, &unitdata, &flag) < 0)
			error ("$*Erro no RCVUDP");

		mutime (&end_time);

		for (sp = servtb; /* abaixo */; sp++)
		{
			if (sp->s_addr == 0)
			{
				error ("Recebendo datagrama UDP de endereo errado");
				goto again;
			}

			if (rcv_addr.a_addr == sp->s_addr)
				break;
		}

		if (vflag) error
		(	"Recebi UDP %d bytes (%s checksum) de \"%s\"",
			unitdata.udata.len, flag & T_NO_CHECKSUM ? "sem" : "com",
			t_addr_to_str (sp->s_addr)
		);

		analyse_udp_datagrams (sp, area);

	}	/* end for (EVER) */

}	/* end pai */
#endif	/*******************************************************/

#if (0)	/*******************************************************/
/*
 ****************************************************************
 *	Analisa a resposta					*
 ****************************************************************
 */
void
analyse_udp_datagrams (SERVTB *sp, char *area)
{
	H		*hp = (H *)area;
	int		i;
	int		len;
	char		*str;
	const char	*cp = area + sizeof (H), *old_cp;
	char		*qname, *aname;
	int		qtype, qclass;
	int		atype, aclass, ttl;
	DNS		dns[2];
#	define	NMAIL	8
	int		n_mail = 0;
	DNS		mail_server[NMAIL];
	DNS		canon;

	/*
	 *	Converte o ENDIAN, se for o caso
	 */
#ifdef	LITTLE_ENDIAN
	hp->h_id    = ENDIAN_SHORT (hp->h_id);
	hp->h_qdcnt = ENDIAN_SHORT (hp->h_qdcnt);
	hp->h_ancnt = ENDIAN_SHORT (hp->h_ancnt);
	hp->h_nscnt = ENDIAN_SHORT (hp->h_nscnt);
	hp->h_arcnt = ENDIAN_SHORT (hp->h_arcnt);
#endif	LITTLE_ENDIAN

	/*
	 *	Pequenas inicializaes
	 */
	area_ptr = area;
	memsetl (dns,    0, sizeof (dns)   / sizeof (long));
	memsetl (&canon, 0, sizeof (canon) / sizeof (long));

	/*
	 *	Analisa o "header"
	 */
#ifdef	DEBUG
	if (vflag)
	{
		printf ("HEADER (id = %d):\n\n", hp->h_id);

		printf (hp->h_qr ? "Response, " : "Query, ");

		printf ("%s, ", h_op_text (hp->h_op));	/* Kind of Query */

		printf (hp->h_aa ? "authoritative, " : "NOT authoritative, ");
		printf (hp->h_tc ? "truncated\n" : "NOT truncated\n");

		printf (hp->h_rd ? "recursion desired, " : "NO recursion desired, ");
		printf (hp->h_ra ? "recursion available, " : "NO recursion available, ");

		printf ("%s\n", h_rc_text (hp->h_rc));	/* Response code */

		printf ("Entries: ");
		printf ("question = %d, ", hp->h_qdcnt);
		printf ("answer = %d, ", hp->h_ancnt);
		printf ("authority = %d, ", hp->h_nscnt);
		printf ("additional = %d\n", hp->h_arcnt);
	}
#endif	DEBUG

	/*
	 *	S deve haver uma pergunta
	 */
	if (hp->h_qdcnt != 1)
	{
		error ("Recebi um datagrama com %d perguntas", hp->h_qdcnt);
		return;
	}

	/*
	 *	Retira a pergunta (ela s  relevante em caso de erro)
	 */
	str    = get_domain_name (&cp); len = strlen (str);
	qname  = alloca (len + 1); strcpy (qname, str);
	qtype  = get_short (&cp);
	qclass = get_short (&cp);

	/*
	 *	Examina o tipo da pergunta
	 */
	if (qclass != CLASS_IN)
	{
		error ("Recebi um datagrama com classe %s", get_class_name (qclass));
		return;
	}

	if (qtype != TYPE_A && qtype != TYPE_MX && qtype != TYPE_PTR)
	{
		error ("Recebi um datagrama com pergunta de tipo %s", get_type_name (qtype));
		return;
	}

	/*
	 *	Se no h respostas,  como se fsse um erro
	 */
	if (hp->h_ancnt == 0 && hp->h_rc == DOMAIN_NO_ERROR)
		hp->h_rc = DOMAIN_NAME_ERROR;

	/*
	 *	Verifica se houve erro
	 */
	if (hp->h_rc != DOMAIN_NO_ERROR)
	{
		int		error_code = DNS_NAME_ERROR;

		sp->s_code = SERVER_OK;

		if (hp->h_rc != DOMAIN_NAME_ERROR)
		{
			error
			(	"Cdigo \"%s\" inesperado do servidor DNS \"%s\"",
				h_rc_text (hp->h_rc), t_addr_to_str (sp->s_addr)
			);
			error_code = DNS_SERVER_ERROR;
			sp->s_code = SERVER_BAD_ANSWER;
		}

	   	if (qtype == TYPE_PTR)
		{
			dns[0].d_host_nm[0] = '\0';
		   	dns[0].d_host_addr = t_arpa_str_to_addr (qname);
		   	dns[0].d_entry_code = 'R';
		}
		else
		{
			strcpy (dns[0].d_host_nm, qname);
		   	dns[0].d_host_addr = 0;
		   	dns[0].d_entry_code = (qtype == TYPE_MX) ? 'M' : 'C';
		}

		dns[0].d_error_code = error_code;
		dns[0].d_server_index = sp->s_index;

		dns[0].d_expir_time = 60;
	   /***	dns[0].d_preference = ...; ***/

		if (ioctl (udp_fd, I_DNS_PUT_INFO, dns) < 0)
			error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");

		return;
	}

	/*
	 *	Processa a seo "answer"
	 */
	for (i = 0; i < hp->h_ancnt; i++, cp = old_cp + len)
	{
		str	= get_domain_name (&cp); len = strlen (str);
		aname	= alloca (len + 1); strcpy (aname, str);
		atype	= get_short (&cp);
		aclass	= get_short (&cp);
		ttl	= get_long (&cp);
		len	= get_short (&cp);

		if (ttl == 0)
			ttl = 8 * 60;

		old_cp = cp;

		/*
		 *	Examina a classe
		 */
		if (aclass != CLASS_IN)
		{
			error ("Recebi uma reposta com classe %s", get_class_name (aclass));
			continue;
		}

		/*
		 *	Examina o tipo da resposta
		 */
		if   (atype == TYPE_PTR)	/* O nome, dado o endereo */
		{
			IPADDR		addr;

			if ((addr = t_arpa_str_to_addr (aname)) == -1)
			{
				error
				(	"Endereo IN.ADDR-ARPA invlido (%s)",
					aname
				);
			}

			strcpy (dns[0].d_host_nm, get_domain_name (&cp));
		   	dns[0].d_host_addr = addr;

		   	dns[0].d_entry_code = 'R';
			dns[0].d_error_code = DNS_NO_ERROR;
			dns[0].d_server_index = sp->s_index;

			dns[0].d_expir_time = ttl;
		   /***	dns[0].d_preference = ...; ***/

			if (ioctl (udp_fd, I_DNS_PUT_INFO, dns) < 0)
				error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");

			continue;
		}
		elif (atype == TYPE_MX)		/* O nome de um servidor de correio */
		{
			const DNS	*dp;
			DNS		*new_dp;

			if (!streq (aname, qname))
			{
				error
				(	"Recebi um servidor de correio esprio (%s, %s)",
					aname, qname
				);
				continue;
			}

			if (n_mail >= NMAIL)
				continue;

			new_dp = &mail_server[n_mail];

			new_dp->d_preference = get_short (&cp);
			strcpy (new_dp->d_host_nm, get_domain_name (&cp));

			/* Processa a preferncia */

		    again:	
			for (dp = mail_server; dp < new_dp; dp++)
			{
				if (dp->d_preference == new_dp->d_preference)
					{ new_dp->d_preference++; goto again; }
			}

			n_mail++;

			/* Envia o servidor de correio (sem endereo IP) */

			strcpy (dns[0].d_host_nm, aname);
		   	dns[0].d_host_addr = 0;

		   	dns[0].d_entry_code = 'M';
			dns[0].d_error_code = DNS_NO_ERROR;
			dns[0].d_server_index = sp->s_index;

			dns[0].d_expir_time = ttl;
		   /***	dns[0].d_preference = ...; ***/


			strcpy (dns[1].d_host_nm, new_dp->d_host_nm);
		   	dns[1].d_host_addr = 0;

		   	dns[1].d_entry_code = 'C';
			dns[1].d_error_code = DNS_NO_ERROR;
			dns[1].d_server_index = sp->s_index;

			dns[1].d_expir_time = ttl;
		   	dns[1].d_preference = new_dp->d_preference;

			if (ioctl (udp_fd, I_DNS_PUT_INFO, dns) < 0)
				error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");

			continue;
		}
		elif (atype == TYPE_CNAME) 	/* O nome cannico de um alias */
		{
			if (!streq (aname, qname))
			{
				error
				(	"Recebi um alias esprio (%s, %s)",
					aname, qname
				);
				continue;
			}

			strcpy (canon.d_host_nm, get_domain_name (&cp));

			continue;
		}
		elif (atype != TYPE_A)
		{
			error
			(	"Recebi um resposta de tipo %s",
				get_type_name (atype)
			);

			continue;
		}

		/*
		 *	 um endereo IP
		 */
		if (streq (aname, canon.d_host_nm))	/* Alias */
		{
			strcpy (dns[0].d_host_nm, qname);
		   	dns[0].d_host_addr = get_long (&cp);

		   	dns[0].d_entry_code = 'A';
			dns[0].d_error_code = DNS_NO_ERROR;
			dns[0].d_server_index = sp->s_index;

			dns[0].d_expir_time = ttl;
		   /***	dns[0].d_preference = ...; ***/


			strcpy (dns[1].d_host_nm, aname);
		   	dns[1].d_host_addr = dns[0].d_host_addr;

		   	dns[1].d_entry_code = 'C';
			dns[1].d_error_code = DNS_NO_ERROR;
			dns[1].d_server_index = sp->s_index;

			dns[1].d_expir_time = ttl;
		   /***	dns[1].d_preference = ...; ***/

			if (ioctl (udp_fd, I_DNS_PUT_INFO, dns) < 0)
				error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");
		}
		else					/* Nome normal */
		{
			strcpy (dns[0].d_host_nm, aname);
		   	dns[0].d_host_addr = get_long (&cp);

		   	dns[0].d_entry_code = 'C';
			dns[0].d_error_code = DNS_NO_ERROR;
			dns[0].d_server_index = sp->s_index;

			dns[0].d_expir_time = ttl;
		   /***	dns[0].d_preference = ...; ***/

			if (ioctl (udp_fd, I_DNS_PUT_INFO, dns) < 0)
				error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");
		}

	}	/* Examinando as respostas */

	/*
	 *	Pula a seo "authority"
	 */
	for (i = 0; i < hp->h_nscnt; i++, cp = old_cp + len)
	{
		(void)get_domain_name (&cp);
		(void)get_short (&cp);
		(void)get_short (&cp);
		(void)get_long (&cp);
		len = get_short (&cp);

		old_cp = cp;

	}	/* Pulando as "autoridades" */

	/*
	 *	Processa a seo "additional"
	 */
	for (i = 0; i < hp->h_arcnt; i++, cp = old_cp + len)
	{
		str	= get_domain_name (&cp); len = strlen (str);
		aname	= alloca (len + 1); strcpy (aname, str);
		atype	= get_short (&cp);
		aclass	= get_short (&cp);
		ttl	= get_long (&cp);
		len	= get_short (&cp);

		if (ttl == 0)
			ttl = 8 * 60;

		old_cp = cp;

		/*
		 *	Examina a classe
		 */
		if (aclass != CLASS_IN)
		{
			error ("Recebi um adicional com classe %s", get_class_name (aclass));
			continue;
		}

		/*
		 *	S considera endereos
		 */
		if (atype != TYPE_A)
			continue;

		/*
		 *	Processa o endereo
		 */
		strcpy (dns[0].d_host_nm, aname);
	   	dns[0].d_host_addr = get_long (&cp);

	   	dns[0].d_entry_code = 'C';
		dns[0].d_error_code = DNS_NO_ERROR;
		dns[0].d_server_index = sp->s_index;

		dns[0].d_expir_time = ttl;
	   /***	dns[0].d_preference = ...; ***/

		if (ioctl (udp_fd, I_DNS_PUT_INFO, dns) < 0)
			error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");

	}	/* Examinando as respostas */

	/*
	 *	Coloca o final dos servidores de correio
	 */
	if (qtype == TYPE_MX)
	{
		strcpy (dns[0].d_host_nm, qname);
	   	dns[0].d_host_addr = 0;

	   	dns[0].d_entry_code = 'M';
		dns[0].d_error_code = DNS_NO_ERROR;
		dns[0].d_server_index = sp->s_index;

		dns[0].d_expir_time = 0;
	   /***	dns[0].d_preference = ...; ***/


		dns[1].d_host_nm[0] = '\0';
	   	dns[1].d_host_addr = 0;

	   	dns[1].d_entry_code = 'E';
		dns[1].d_error_code = DNS_NO_ERROR;
		dns[1].d_server_index = sp->s_index;

		dns[1].d_expir_time = 0;
	   	dns[1].d_preference = 0;

		if (ioctl (udp_fd, I_DNS_PUT_INFO, dns) < 0)
			error ("*Erro no \"ioctl\" I_DNS_PUT_INFO");
	}

}	/* end analyse_udp_datagrams */
#endif	/*******************************************************/

#if (0)	/*******************************************************/
/*
 ****************************************************************
 *	Obtm um nome de domnio				*
 ****************************************************************
 */
char *
get_domain_name (const char **cpp)
{
	const char	*cp = *cpp;
	int		i;
	static char	area[256];
	char		*ap = area;
	char		updated_ptr = 0;

	/*
	 *	Retira o nome
	 */
	while (*cp != 0)
	{
		if ((*cp & 0xC0) == 0xC0)
		{
			if (!updated_ptr)
				{ *cpp = cp + 2; updated_ptr = 1; }

			cp = area_ptr + (((cp[0] & ~0xC0) << 8) | cp[1]);
		}

		for (i = *cp++; i > 0; i--)
			*ap++ = *cp++;

		if (*cp != 0)
			*ap++ = '.';
	}

	*ap++ = '\0';

	if (!updated_ptr)
		*cpp = cp + 1;

	return (area);

}	/* end get_domain_name */

/*
 ****************************************************************
 *	Compe um nome no formato do "domain"			*
 ****************************************************************
 */
void
put_domain_name (char *name, char **area)
{
	char		*np = name;
	char		c;
	int		n = 0;
	char		*ap = *area + 1;
	char		*bp = *area;

	for (EVER)
	{
		*ap++ = c = *np++;

		if (c == '.' || c == '\0')
		{
			*bp = n; n = 0; bp = ap - 1;

			if (c == '\0')
				{ *area = ap; return; }
		}
		else
		{
			n++;
		}

	}	/* end for (EVER) */

}	/* end put_domain_name */

/*
 ****************************************************************
 *	Obtm um "short" desalinhado				*
 ****************************************************************
 */
int
get_short (const char **cpp)
{
	int		n;
	const char	*cp = *cpp;

	n  = (*cp++ << 8);
	n |= *cp++;

	*cpp = cp;

	return (n);

}	/* end get_short */

/*
 ****************************************************************
 *	Coloca um "short" desalinhado				*
 ****************************************************************
 */
void
put_short (int value, char **cpp)
{
	char		*cp = *cpp;

	*cp++ = value >> 8;
	*cp++ = value;

	*cpp = cp;

}	/* end put_short */

/*
 ****************************************************************
 *	Obtm um "long" desalinhado				*
 ****************************************************************
 */
long
get_long (const char **cpp)
{
	long		n;
	const char	*cp = *cpp;

	n  = (*cp++ << 24);
	n |= (*cp++ << 16);
	n |= (*cp++ << 8);
	n |= *cp++;

	*cpp = cp;

	return (n);

}	/* end get_long */

/*
 ****************************************************************
 *	Decodifica um nome reverso  ("d.d.d.d.IN-ADDR.ARPA")	*
 ****************************************************************
 */
IPADDR
t_arpa_str_to_addr (const char *node_str)
{
	register long		i, n;
	register const char	*cp = node_str;
	const char		*str;
	char			node[4];
	register char		*np;

#ifdef	LITTLE_ENDIAN
	np = &node[0];
#else
	np = &node[4];
#endif	LITTLE_ENDIAN

	for (i = 0; i < 4; i++)
	{
		if ((n = strtol (cp, &str, 0)) & ~0xFF || *str != '.' || str == cp)
			return (-1);

		cp = str + 1;

#ifdef	LITTLE_ENDIAN
		*np++ = n;
#else
		*--np = n;
#endif	LITTLE_ENDIAN
	}

	return (*(IPADDR *)node);

}	/* end t_arpa_str_to_addr */
#ifdef	DEBUG
/*
 ****************************************************************
 *	Decodifica a operao do cabealho			*
 ****************************************************************
 */
char *
h_op_text (int op)
{
	switch (op)
	{
	    case 0:
		return ("standard query");

	    case 1:
		return ("inverse query");

	    case 2:
		return ("server status request");
	}

	return ("reserved");

}	/* end h_op_text */
#endif	DEBUG

/*
 ****************************************************************
 *	Obtm a descrio do TIPO				*
 ****************************************************************
 */
const char *
get_type_name (int type)
{
	static char	area[24];

	if ((unsigned)type >= sizeof (type_desc) / sizeof (TYPE_DESC) - 1)
	{
		sprintf (area, "Unknown type: %d", type);
		return (area);
	}

	return (type_desc[type].t_name);

}	/* end get_type_name */

/*
 ****************************************************************
 *	Obtm a descrio da CLASSE				*
 ****************************************************************
 */
const char *
get_class_name (int class)
{
	static char	area[24];

	if ((unsigned)class >= sizeof (class_desc) / sizeof (TYPE_DESC) - 1)
	{
		sprintf (area, "Unknown class: %d", class);
		return (area);
	}

	return (class_desc[class].c_name);

}	/* end get_class_name */

/*
 ****************************************************************
 *	Decodifica o cdigo de retorno do servidor		*
 ****************************************************************
 */
char *
h_rc_text (enum DOMAIN_ERROR_CODE code)
{
	switch (code)
	{
	    case DOMAIN_NO_ERROR:
		return ("DOMAIN_NO_ERROR");

	    case DOMAIN_FORMAT_ERROR:
		return ("DOMAIN_FORMAT_ERROR");

	    case DOMAIN_SERVER_FAILURE:
		return ("DOMAIN_SERVER_FAILURE");

	    case DOMAIN_NAME_ERROR:
		return ("DOMAIN_NAME_ERROR");

	    case DOMAIN_NOT_IMPLEMENTED:
		return ("DOMAIN_NOT_IMPLEMENTED");

	    case DOMAIN_REFUSED:
		return ("DOMAIN_REFUSED");
	}

	return ("DOMAIN_???");

}	/* end h_rc_text */

#ifdef	DEBUG
/*
 ****************************************************************
 *	Decodifica o cdigo de retorno para o KERNEL		*
 ****************************************************************
 */
char *
edit_kernel_code (enum DNS_ERROR_CODE code)
{
	switch (code)
	{
	    case DNS_NO_ERROR:
		return ("DNS_NO_ERROR");

	    case DNS_NAME_ERROR:
		return ("DNS_NAME_ERROR");

	    case DNS_SERVER_ERROR:
		return ("DNS_SERVER_ERROR");
	}

	return ("DNS_???");

}	/* end edit_kernel_code */
#endif	DEBUG

/*
 ****************************************************************
 *	Em caso de sinal, ...					*
 ****************************************************************
 */
void
on_signal (int signo, ...)
{
	kill (pidfilho, SIGHUP);

	exit (0);

}	/* on_signal */
#endif	/*******************************************************/

/*
 ****************************************************************
 *	Imprime uma mensagem de erro (servidor)			*
 ****************************************************************
 */
void
error (const char *format, ...)
{
	va_list		args;
	const char	*fmt = format;
	FILE		*log_fp = NOFILE;
	char		err, abrt;
	int		event = 0, save_errno = errno;
	time_t		time_loc;
	static int	static_log_fd = 0;
	static FILE	*static_log_fp = NOFILE;

	va_start (args, format);

	/*
	 *	Procura o arquivo de "log"
	 */
	if   (error_msg_to_log == 0)
	{
		log_fp = stderr;
	}
	elif (static_log_fd == 0)
	{
		if ((static_log_fd = open ("/var/adm/itnetlog", O_WRONLY|O_APPEND)) >= 0)
		{
			if ((static_log_fp = log_fp = fdopen (static_log_fd, "w")) == NOFILE)
				static_log_fd = -1;
		}
	}
	else
	{
		log_fp = static_log_fp;
	}

	/*
	 *	Processa os argumentos da cadeia
	 */
	if (fmt == NOSTR)
		fmt = "$*Erro de entrada/sada de TCP";

	if ((abrt = *fmt) == '$')
		fmt++;

	if ((err = *fmt) == '*')
	{
		fmt++;

		if (save_errno == TLOOK)
			event = t_look (udp_fd);
	}

	/*
	 *	Escreve as mensagens
	 */
	if (log_fp != NOFILE)
	{
		/* Escreve a mensagem com o tempo */

		fprintf  (log_fp, pgname);

		time (&time_loc);
		fprintf (log_fp, " (%.24s)", btime (&time_loc));

		fprintf  (log_fp, ": ", pgname);
		vfprintf (log_fp, fmt, args);

		/* Escreve o tipo do erro */

		if (err == '*')
		{
			if (event != 0)
				fprintf (log_fp, " (%s)", t_strevent (udp_fd, event));
			elif (save_errno != 0)
				fprintf (log_fp, " (%s)", strerror (save_errno));
		}

		putc ('\n', log_fp); fflush (log_fp);
	}

	if (abrt == '$' || event ==  T_DISCONNECT)
		exit (1);

	va_end (args);

}	/* end error */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
		"%s - consulta um servidor de DNS para obter um endereo\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"\t%s [-v]\n",
		pgname, pgversion, pgname
	);
	fprintf
	(	stderr,
		"\nOpes:"
		"\t-v: Verboso\n"
	);

	exit (2);

}	/* end help */
