/*
 ****************************************************************
 *								*
 *			main.c					*
 *								*
 *	File transfer protocol (servidor)			*
 *								*
 *	Verso	3.0.0, de 12.10.92				*
 *		4.0.0, de 11.08.01				*
 *								*
 *	Mdulo: FTP_S						*
 *		Utilitrios de comunicaes			*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2001 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/syscall.h>
#include <sys/utsname.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <xti.h>

#include "ftp_s.h"
#include "../ftp.h"

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
#define	CNTL_TIME_OUT	600	/* Tempo de espera por um comando */

const char	pgversion[] =  "Verso:	4.0.0, de 11.08.01";

const char	tcp_dev_nm[] = "/dev/itntcp";

entry int	vflag;		/* Verbose */

entry int	type = TYPE_A;	/* Tipo corrente */
entry int	form = FORM_N;	/* Forma corrente */
entry int	stru = STRU_F;	/* Tipo corrente */
entry int	mode = MODE_S;	/* Modo corrente */

entry int	cntl_fd;	/* Conexo de controle */

entry FILE	*cntl_fp;	/* Conexo de controle */

entry INADDR	server_cntl_addr[2];	/* Endereo do servidor */

entry char	client_node_nm[80];	/* Nome do cliente */

entry int	snd_dis;	/* Em caso de erro grave envia DIS */

entry int	global_log_fd;	/* Para "error" */
entry FILE	*global_log_fp;

/*
 ****************************************************************
 *	Programa Principal					*
 ****************************************************************
 */
int
main (int argc, register const char *argv[])
{
	register int	opt, n;
	const char	*str;
	UTSNAME		uts;
	INADDR		client_addr;
	INADDR		serv_addr[2];
	T_BIND		bind;
	T_CALL		call;
	T_OPTMGMT	req_optmgmt, ret_optmgmt;
	TCP_OPTIONS	options;
	int		listen_port, listen_fd, sequence;

	/*
	 *	Prepara variveis de "error" ainda como superusurio
	 */
	if ((global_log_fd = open ("/var/adm/itnetlog", O_WRONLY|O_APPEND)) >= 0)
	{
		if ((global_log_fp = fdopen (global_log_fd, "w")) == NOFILE)
			global_log_fd = -1;
	}

	/*
	 *	Analisa as opes
	 */
	while ((opt = getopt (argc, argv, "v")) != EOF)
	{
	   	switch (opt)
		{
		    case 'v':			/* Verbose */
			 vflag++;
			 break;

		}	/* end switch */

	}	/* end while */

	argv += optind;
	argc -= optind;

	/*
	 *	Obtm o UTS do sistema
	 */
	if (uname (&uts) < 0)
		error ("$*No consegui obter o UTS\n");

	/*
	 *  tcmpto_s [-v] <listen_port> <listen_fd>
	 *				<client_addr> <client_port> <sequence>
	 */
#ifdef	DEBUG
	if (Bflag) error
	(	"%s: %s <%s> <%s> <%s> <%s> <%s>",
		pgname, vflag ? "-v" : "",
		argv[0], argv[1], argv[2], argv[3], argv[4]
	);
#endif	DEBUG

	if (argc != 5) error
	(	"$%s: Nmero invlido de argumentos: %d",
		pgname, argc
	);

	listen_port	   = strtol (argv[0], &str, 0);
	listen_fd	   = strtol (argv[1], &str, 0);
	client_addr.a_addr = t_str_to_addr (argv[2]);
	client_addr.a_port = strtol (argv[3], &str, 0);
	sequence	   = strtol (argv[4], &str, 0);

	if ((cntl_fd = t_open (tcp_dev_nm, O_RDWR, (T_INFO *)NULL)) < 0)
		error ("$*No consegui abrir \"%s\"", tcp_dev_nm);

	if ((cntl_fp = fdopen (cntl_fd, "w")) == NOFILE)
		error ("$*No consegui obter um FILE para \"%s\"", tcp_dev_nm);

	/*
	 *	Vou realizar o "bind"
	 */
	serv_addr[0].a_port = listen_port;	/* Usa o FTP_CNTL_PORT */
	serv_addr[1].a_addr = client_addr.a_addr;
	serv_addr[1].a_port = client_addr.a_port;

	bind.addr.maxlen = 0;
	bind.addr.len	 = 2 * sizeof (INADDR);
	bind.addr.buf	 = serv_addr;

	bind.qlen	 = 0;

	if (t_bind (cntl_fd, &bind, (T_BIND *)NULL) < 0)
		error ("$*No consegui atribuir endereo");

	/*
	 *	Desliga o "timeout" do TCP/IP
	 */
	req_optmgmt.flags = T_DEFAULT;

   /***	req_optmgmt.opt.buf    = (TCP_OPTIONS *)NULL; ***/
   /***	req_optmgmt.opt.len    = 0; ***/

	ret_optmgmt.opt.buf    = &options;
	ret_optmgmt.opt.maxlen = sizeof (TCP_OPTIONS);

	if (t_optmgmt (cntl_fd, &req_optmgmt, &ret_optmgmt) < 0)
		error ("$*No consegui negociar (DEFAULT)");

	req_optmgmt.flags = T_NEGOTIATE;

	options.max_silence = 0;

	req_optmgmt.opt.buf    = &options;
	req_optmgmt.opt.len    = sizeof (TCP_OPTIONS);

   /***	ret_optmgmt.opt.buf    = &options; ***/
   /***	ret_optmgmt.opt.maxlen = sizeof (TCP_OPTIONS); ***/

	if (t_optmgmt (cntl_fd, &req_optmgmt, &ret_optmgmt) < 0)
		error ("$*No consegui negociar (NEGOCIATE)");

	/*
	 *	Vou realizar o "accept"
	 */
   /***	call.opt.len = 0; ***/
   /***	call.udata.len = 0; ***/
	call.sequence = sequence;

	call.addr.len	 = sizeof (INADDR);
	call.addr.maxlen = sizeof (INADDR);
	call.addr.buf	 = &client_addr;

#ifdef	DEBUG
	if (Bflag) error
	(	"Vou dar accept para (%s, %d)",
		t_addr_to_str (client_addr.a_addr), client_addr.a_port
	);
#endif	DEBUG

	if (t_accept (listen_fd, cntl_fd, &call) < 0)
		error ("$*Erro no t_accept");

	/*
	 *	Agora no precisa mais do "listen_fd"
	 */
	client_data_addr.a_addr = client_addr.a_addr;

	t_close (listen_fd);

	if ((str = t_addr_to_node (cntl_fd, client_addr.a_addr)) == NOSTR)
		str = t_addr_to_str (client_addr.a_addr);

	strcpy (client_node_nm, str);

	/*
	 *	Obtm o endereo local do servidor
	 */
	if (t_getaddr (cntl_fd, server_cntl_addr) < 0)
		error ("$*No consegui obter o endereo local");

	reply
	(	220, "%s (%s, %s) Servidor FTP (%s)",
		uts.uts_sysname, uts.uts_machine, uts.uts_nodename, pgversion
	);

	/*
	 *	L os comandos do cliente e processa
	 */
	for (EVER)
	{
		char	comando[BLSZ];

		alarm (CNTL_TIME_OUT);

		if ((n = t_rcv (cntl_fd, comando, sizeof (comando), (int *)NULL)) < 0)
			error (NOSTR);

		alarm (0);
#undef	DEBUG
#ifdef	DEBUG
		{
			char			line[BLSZ];
			register const char	*ip = comando;
			register char		*op = line;
			register char		c;
			register int		i;

			for (i = n; i > 0; i--)
			{
				*op++ = c = *ip++;

				if (c == '\r' || c == '\n' || c == '\0')
					{ op--; break; }
			}

			*op = '\0';

			error ("Recebi \"%s\"", line);
		}
#endif	DEBUG
#undef	DEBUG
		if (n == 0)
			strcpy (comando, "QUIT\r\n");
		else
			comando[n] = '\0';	/* Para marcar o final da linha */

		parse (comando);
	}

}	/* end main */

/*
 ****************************************************************
 *	Analisador de  Comandos					*
 ****************************************************************
 */
void
parse (const char *comando)
{
	register const CMDTB	*tp;
	register const char	*cp;
	register int		argc;
	const char		*argv[10];

	/*
	 *	Procura o comando na tabela 
	 */
	cp = strtoupper (strtok ((char *)comando, " \t\r\n"));

	for (tp = cmdtab; /* teste abaixo */; tp++)
	{
		if (tp->c_nome == NOSTR)
		{
			reply (500, "Comando \"%s\" desconhecido", cp);
			return;
		}

		if (streq (cp, tp->c_nome))
			break;
	}

	if (tp->c_funcao == NOFUNC)
	{
		reply (502, "Comando \"%s\" no implementado", cp);
		return;
	}

	/*
	 *	Retira os argumentos
	 */
	for (argc = 0; (cp = strtok (NOSTR, " \t\r\n")) != NOSTR; argc++)
		argv[argc] = cp;

	argv[argc] = NOSTR;

#ifdef	DEBUG
	{
		register int i;

		printf ("cmd_parse: (%d) \"%s", argc, tp->c_nome);

		for (i = 0; i < argc; i++)
			printf (" %s", argv[i]);

		printf ("\"\n");
	}
#endif	DEBUG

	/*
	 *	Executa a funo pedida
	 */
	(*tp->c_funcao) (argc, argv);

}	/* end parse */

/*
 ****************************************************************
 *	Verifica se est no estado READY			*
 ****************************************************************
 */
int
check_ready_state (void)
{
	if (server_state == S_READY)
		return (0);

	reply (503, "Este comando s pode ser executado com um usurio em sesso");

	return (-1);

}	/* end check_ready_state */

/*
 ****************************************************************
 *	Converte uma cadeia para maisculas			*
 ****************************************************************
 */
char *
strtoupper (char *str)
{
	register char *cp = str;

	if (cp == NOSTR)
		return (NOSTR);

	for (/* vazio */; *cp != '\0'; cp++)
		*cp = toupper (*cp);

	return (str);

}	/* end strtoupper */

/*
 ****************************************************************
 *	Envia uma mensagem longa para o cliente			*
 ****************************************************************
 */
void
long_reply (int code, const char *format, ...)
{
	register va_list	args;
	char			msg[BLSZ];

	va_start (args, format);

	sprintf  (msg, "%03d-", code);
	vsprintf (msg + 4, format, args);
	strcat (msg, "\r\n");

	if (t_snd (cntl_fd, msg, strlen (msg), T_PUSH) < 0)
		error (NOSTR);

	va_end (args);

}	/* end long_reply */

/*
 ****************************************************************
 *	Envia uma mensagem normal para o cliente		*
 ****************************************************************
 */
void
reply (int code, const char *format, ...)
{
	register va_list	args;
	char			msg[BLSZ];

	va_start (args, format);

	sprintf  (msg, "%03d ", code);
	vsprintf (msg + 4, format, args);
	strcat (msg, "\r\n");

	if (t_snd (cntl_fd, msg, strlen (msg), T_PUSH) < 0)
		error (NOSTR);

	va_end (args);

}	/* end reply */

/*
 ****************************************************************
 *	Imprime uma mensagem de erro (servidor)			*
 ****************************************************************
 */
void
error (const char *format, ...)
{
	register va_list	 args;
	register const char	*fmt = format;
	register FILE		*log_fp;
	char			err, abrt;
	int			event = 0, save_errno = errno;
	time_t			time_loc;
	static int		static_log_fd = 0;
	static FILE		*static_log_fp = NOFILE;

	va_start (args, format);

	/*
	 *	Procura o arquivo de "log"
	 */
	if (static_log_fd == 0)
	{
		if ((static_log_fd = open ("/var/adm/itnetlog", O_WRONLY|O_APPEND)) >= 0)
		{
			if ((static_log_fp = fdopen (static_log_fd, "w")) == NOFILE)
				static_log_fd = -1;
		}
	}

	log_fp = static_log_fp;

	/*
	 *	Processa os argumentos da cadeia
	 */
	if (fmt == NOSTR)
		fmt = "$*Erro de entrada/sada de TCP";

	if ((abrt = *fmt) == '$')
		fmt++;

	if ((err = *fmt) == '*')
	{
		fmt++;

		if (save_errno == TLOOK)
			event = t_look (cntl_fd);
	}

	/*
	 *	Escreve as mensagens
	 */
	if (log_fp != NOFILE)
	{
		/* Escreve a mensagem com o tempo */

		fprintf  (log_fp, pgname);

		time (&time_loc);
		fprintf (log_fp, " (%.24s)", btime (&time_loc));

		fprintf  (log_fp, ": ", pgname);
		vfprintf (log_fp, fmt, args);

		/* Escreve o tipo do erro */

		if (err == '*')
		{
			if   (event != 0)
				fprintf (log_fp, " (%s)", t_strevent (cntl_fd, event));
			elif (save_errno != 0)
				fprintf (log_fp, " (%s)", strerror (save_errno));
		}

		putc ('\n', log_fp); fflush (log_fp);
	}

	if (abrt == '$' || event ==  T_DISCONNECT)
		tcp_abort (1);

	va_end (args);

}	/* end error */

/*
 ****************************************************************
 *	Encerra as atividades (desconectando o TCP)		*
 ****************************************************************
 */
void
tcp_abort (int n)
{
	if (snd_dis)
	{
		t_snddis (cntl_fd, (T_CALL *)NULL);

		if (data_fd > 0)
			t_snddis (data_fd, (T_CALL *)NULL);
	}

	exit (n);

}	/* end tcp_abort */
