/*
 ****************************************************************
 *								*
 *			ld.c					*
 *								*
 *	Link-editor de mdulos objeto				*
 *								*
 *	Verso	1.0.0, de 08.01.87				*
 *		4.3.0, de 12.07.02				*
 *								*
 *	Mdulo: ld						*
 *		Utilitrios bsicos				*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2002 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <a.out.h>

#include "../h/ld.h"

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
const char		pgversion[] =  "Verso:	4.3.0, de 12.07.02";

/*
 ****** Opes de link-edio ***********************************
 */
entry int		rflag,		/* Mantem a relocao, no define common */
			bflag,		/* Gera uma biblioteca compartilhada */
			nflag,		/* mdulo executavel reentrante */
			sflag,		/* "Strip" - Remove SYMTB, RT e RD */
			vflag,		/* Verbose */
			mflag,		/* Gera mapa de link-edio */
			uflag,		/* lista os no referenciados */
			dotflag;	/* Temporrias no diretrio corrente */

/*
 ****** Variveis globais ***************************************
 */
entry int		shared_lib_given; /* Foram dadas bibliotecas compartilhadas */

entry const char	*obj_nm = "a.out"; /* Nome do mdulo final */

entry const char	*entry_nm;	/* Nome do smbolo de "Entry-Point" */

entry off_t		TEXT_ORG = -1,	/* Endereco inicial do "TEXT" */
			DATA_ORG = -1,	/* Endereco inicial do "DATA" */
			COMMON_ORG,	/* Endereco inicial do "COMMON" */
			BSS_ORG;	/* Endereco inicial do "BSS" */

/*
 ******	Prottipos de funes ***********************************
 */
void			on_signal (int, ...);
void	 		quit (int);
void			help (void);

/*
 ****************************************************************
 *	Link-editor de mdulos objeto				*
 ****************************************************************
 */
int
main (int argc, const char *argv[])
{
	int		opt;
	const char	*str;

	/*
	 *	Analisa a linha de argumentos para o programa
	 *
	 *	ld [-rbsmnvu.] [-o <obj>] [-e <ent>] [-T <tb>] [-D <db>] <m> ... [-l<x> ...]
	 *
	 */
	while ((opt = getopt (argc, argv, "rbsmnvu.o:e:T:D:MH")) != EOF)
	{
		switch (opt)
		{
		    case 'r':			/* Mantenha a relocao */
			rflag++;
			break;

		    case 'b':			/* Gera uma biblioteca compartilhada */
			bflag++;
			break;

		    case 's':			/* Sem tabela de smbolos */
			sflag++;
			break;

		    case 'm':			/* Imprime o mapa de mdulos */
			mflag++;
			break;

		    case 'n':			/* Programa reentrante */
			nflag++;
			break;

		    case 'v':			/* Verbose */
			vflag++;
			break;

		    case 'u':			/* Lista os no referenciados */
			uflag++;
			break;

		    case '.':			/* Temporrias no diretrio corrente */
			dotflag++;
			break;

		    case 'o':			/* Nome do objeto gerado */
			obj_nm = optarg;
			break;

		    case 'e':			/* Nome do ponto de entrada */
			entry_nm = optarg;
			break;

		    case 'T':			/* Origem da seo TEXT */
			TEXT_ORG = strtol (optarg, &str, 0);

			if (*str != '\0')
				error ("@Origem da seo TEXT \"%s\" invlida", optarg);

			break;

		    case 'D':			/* Origem da seo DATA */
			DATA_ORG = strtol (optarg, &str, 0);

			if (*str != '\0')
				error ("@Origem da seo DATA \"%s\" invlida", optarg);

			break;

		    case 'M':
			exit(0);

		    case 'H':		/* Help */
			help ();

		    default:		/* Erro */
			fprintf (stderr, "\n");
			help ();

		}	/* end switch */

	}	/* end while */

	if (nflag + rflag + bflag > 1)
		error ("@Opes conflitantes: \"-n\", \"-r\" e \"-b\" so mutuamente exclusivas");

	if ((rflag + bflag) && sflag)
		error ("@Opes conflitantes: \"-r\" ou \"-b\" com \"-s\"");

	if (error_standard_count)
		quit (1);

   /***	argc -= optind; ***/
	argv += optind;

	if (*argv == NOSTR)
		return (0);

	/*
	 *	Programa os sinais INT e QUIT
	 */
	if (signal (SIGINT,  SIG_IGN) != SIG_IGN)
		signal (SIGINT, on_signal);

	if (signal (SIGQUIT, SIG_IGN) != SIG_IGN)
		signal (SIGQUIT, on_signal);

	/*
	 *	Prepara "error"
	 */
	error_default_function = quit;

	/*
	 ******	Executa os diversas passos ******************************
	 */

	/*
	 *	Passo 1: Monta a Tabela Global de Smbolos
	 */
	pass1 (argv);

	if (error_standard_count)
		quit (1);

	/*
	 *	Passo 2: Verifica os mdulos necessrios das bibliotecas regulares
	 */
	pass2 ();

	if (error_standard_count)
		quit (1);

	/*
	 *	Passo 3: Atribui valores definitivos aos smbolos
	 */
	pass3 ();

	if (error_standard_count)
		quit (1);

	/*
	 *	Passo 4: Prepara os arquivos para a gerao
	 */
	pass4 ();

	if (error_standard_count)
		quit (1);

	/*
	 *	Passo 5: Gera as diversas sees
	 */
	pass5 ();

	if (error_standard_count)
		quit (1);

	/*
	 *	Passo 6: Verifica os smbolos das bibliotecas compartilhadas
	 */
	pass6 ();

	if (error_standard_count)
		quit (1);

	/*
	 *	Passo 7: Coda: Escreve as diversas sees e o cabealho
	 */
	pass7 ();

	if (error_standard_count)
		quit (1);

#ifdef	COLISO
	colision_tst ();
#endif	COLISO

	return (0);

}	/* end ld */

/*
 ****************************************************************
 *	Em caso de sinal, ...					*
 ****************************************************************
 */
void
on_signal (int signo, ...)
{
	quit (1);

}	/* end on_signal */

/*
 ****************************************************************
 *	Termina o Programa Anormalmente				*
 ****************************************************************
 */
void
quit (int n)
{
	if (aout_fd >= 0)
		unlink (obj_nm);

   /***	unlink (text_nm); ***/
   /***	unlink (data_nm); ***/

	if (rt_nm)
		unlink (rt_nm);

	if (rd_nm)
		unlink (rd_nm);

	exit (n);

}	/* end quit */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
		"%s - link-editor de mdulos objeto\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"  %s [-rbsmnvu.] [-o <obj>] [-e <ent>] "
		"[-T <tb>] [-D <db>] <m> ... [-l<x> ...]\n",
		pgname, pgversion, pgname
	);

	fprintf
	(	stderr,
		"\nOpes:"
		"\t-r: Mantem a relocao, no define COMMON\n"
		"\t-b: Gera uma biblioteca compartilhada\n"
		"\t-s: Gera o mdulo objeto sem tabela de smbolos\n"
		"\t-m: Gera um mapa de link-edio\n"
		"\t-n: Gera o mdulo objeto reentrante\n"
		"\t-v: Verboso\n"
		"\t-u: Lista os smbolos no referenciados\n"
		"\t    Se for dado duas vezes, lista tambm os "
			"das bibliotecas\n"
		"\t-.: Usa os arquivos temporrios no diretrio corrente\n"
	);

	fprintf
	(	stderr,
		"\t-o: O nome do mdulo gerado ser <obj> ao invs de \"a.out\"\n"
		"\t-e: O nome do smbolo de entrada ser <ent> ao invs de \"start\"\n"
		"\t-T: O endereo do incio da seo \"text\" ser <tb>\n"
		"\t-D: O endereo do incio da seo \"data\" ser <db>\n"
		"\nObs.:"
		"\t<m> ... so mdulos objeto\n"
		"\t<x> ... so abreviaturas para nomes de bibliotecas\n"
	);

	exit (2);

}	/* end help */
