/*
 ****************************************************************
 *								*
 *			pass5.c					*
 *								*
 *	Passo 5: Gera as diversas sees			*
 *								*
 *	Verso	1.0.0, de 23.10.86				*
 *		4.3.0, de 10.06.02				*
 *								*
 *	Mdulo: ld						*
 *		Utilitrios bsicos				*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2002 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <a.out.h>

#include "../h/ld.h"

/*
 ****************************************************************
 *	Definies globais					*
 ****************************************************************
 */
#define	REF_VEC_INCR	32		/* Incrementos para o vetor de referncias */

/*
 ****** Variveis globais ***************************************
 */
entry const char 	*text_nm,	/* Nome do arquivo com as sees "TEXT" */
			*data_nm,	/* Nome do arquivo com as sees "DATA" */
			*rt_nm,		/* Nome do arquivo com a RT */
			*rd_nm;		/* Nome do arquivo com a RD */

entry int		aout_fd = -1,	/* Descritor do mdulo final */
			text_fd,	/* Descritor do arquivo com as sees "TEXT" */
			data_fd;	/* Descritor do arquivo com as sees "DATA" */

entry FILE	 	*rt_fp,		/* Descritor do arquivo com a RT */
			*rd_fp;		/* Descritor do arquivo com a RD */

entry off_t		TEXT_REL_SIZE,	/* Tamanho da RT */
			DATA_REL_SIZE;	/* Tamanho da RD */

entry off_t		text_addr,	/* Endereco do "TEXT" do mdulo atual */
			data_addr,	/* Endereco do "DATA" do mdulo atual */
			bss_addr;	/* Endereco do "BSS"  do mdulo atual */

entry off_t		text_rel,	/* Relocao para o "TEXT" */
			data_rel,	/* Relocao para o "DATA" */
			bss_rel;	/* Relocao para o "BSS"  */

entry int		external_reference_cnt; /* No. de smbolos externos */

/*
 ******	Prottipos de funes ***********************************
 */
void		generate_module_sections (const MODULE *);
void 		relocate
		(	const char *,
			int,
			const char *,
			unsigned,
			off_t,
			unsigned,
			off_t,
			GSYM * const *,
			int,
			int,

			FILE *,
			off_t *,
			off_t *,

			off_t
		);
void		insert_external_reference (GSYM *gp, off_t offset);

/*
 ****************************************************************
 *	Passo 5: Gera as diversas sees			*
 ****************************************************************
 */
void
pass5 (void)
{
	const MODULE	*mp;

	/*
	 *	Prepara os endereo inicial
	 */
	text_addr = TEXT_ORG;
	data_addr = DATA_ORG;
	bss_addr  = BSS_ORG;

	if (mflag)
	{
		if (vflag)
			printf ("\n\n");

		/*
		 *	Imprime o Cabealho do Mapa
		 */
		printf
		( 	"---- TEXTO ----     "
			"---- DATA -----     "
			"----- BSS -----\n"
			"ENDEREO  TAMAN     "
			"ENDEREO  TAMAN     "
			"ENDEREO  TAMAN     "
			"--- MDULO ---  "
			"\n\n"
		);

		vflag = 0;
	}

	/*
	 *	Percorre a lista de mdulos
	 */
	for (mp = modtb_list.m_next; mp != NOMODULE; mp = mp->m_next)
	{
		if (vflag)
			printf ("Mdulo: %s\n", mp->m_nm);

		if (mp->m_type == FMAGIC)
			generate_module_sections (mp);
	}

	/*
	 *	Imprime os totais
	 */
	if (mflag)
	{
		printf
		(	"\nTOTAIS: Text = %d, Data = %d, ",
			TEXT_SIZE, DATA_SIZE
		);
		printf
		(	"Bss = %d, Common = %d\n\n",
			BSS_SIZE, COMMON_SIZE
		);
	}

}	/* end pass5 */

/*
 ****************************************************************
 *	Carrega um mdulo objeto				*
 ****************************************************************
 */
void
generate_module_sections (const MODULE *mp)
{
	const char	*nm = mp->m_nm;
	SYM		*sp;
	GSYM		**vp, *gp;
	off_t		rt_offset;
	GSYM		**vector_image = (GSYM **)NULL;
	int		vector_size, vector_elem = 0;
	static off_t	text_offset, data_offset;
	int		mod_fd;
	HEADER		h;
	void		*symtb_image = NOSYM;
	const SYM	*end_symtb_image = NOSYM;

	/*
	 *	Abre o mdulo e le o seu cabealho
	 */
	if ((mod_fd = inopen (mp->m_dev, mp->m_ino)) < 0)
		{ error ("*@No consegui abrir \"%s\"", nm); return; }

	if (read (mod_fd, &h, sizeof (HEADER)) != sizeof (HEADER))
		{ error ("*@Erro na leitura do cabealho de \"%s\"", nm); close (mod_fd); return; }

	if (h.h_magic != FMAGIC)
		{ error ("@O mdulo \"%s\" NO  regular", nm); close (mod_fd); return; }

	/*
	 *	Calcula as constantes de relocao
	 */
	text_rel  = text_addr;
	data_rel  = data_addr - h.h_tsize;
	bss_rel   = bss_addr  - h.h_tsize - h.h_dsize;

	/*
	 *	Imprime a linha da tabela de mdulos
	 */
	if (mflag)
	{
		printf
		(	"%P %5d     %P %5d     %P %5d     %s\n",
			text_addr, h.h_tsize,
			data_addr, h.h_dsize,
			bss_addr,  h.h_bsize,
			nm
		);
	}

	/*
	 *	Processa a tabela de smbolos
	 */
	if (h.h_ssize != 0)
	{
		/*
		 *	Le a tabela de smbolos
		 */
		symtb_image = alloca (h.h_ssize);

		lseek (mod_fd, h.h_tsize + h.h_dsize, SEEK_CUR);

		if (read (mod_fd, symtb_image, h.h_ssize) != h.h_ssize)
		{
			error ("*@Erro na leitura da tabela de smbolos de \"%s\"", nm);
			close (mod_fd); return;
		}

		end_symtb_image = symtb_image + h.h_ssize;

		/*
		 *	Aloca e zera uma rea para o vetor de ponteiros para a SYMTB
		 */
		for (sp = symtb_image; sp < end_symtb_image; sp = SYM_NEXT_PTR (sp))
			vector_elem++;

		vector_size = vector_elem * sizeof (GSYM *);

		vector_image = vp = alloca (vector_size);

		memsetl (vector_image, 0, vector_size / sizeof (long));

		/*
		 *	Processa os Smbolos
		 */
		for (sp = symtb_image; sp < end_symtb_image; sp = SYM_NEXT_PTR (sp), vp++)
		{
			/*
			 *	Escolhe os smbolos significativos
			 */
			switch (sp->s_type)
			{
			    case EXTERN|TEXT:
			    case EXTERN|DATA:
			    case EXTERN|BSS:
			    case EXTERN|ABS:
			    case EXTERN|UNDEF:
				break;

			    default:		/* Ignora os outros smbolos */
				continue;

			}	/* end switch */

			/*
			 *	Obtm a entrada da SYMTB global
			 */
			if ((gp = hash (sp->s_name, sp->s_nm_len)) != NOGSYM)
				{ *vp = gp; continue; }

			error ("@ERRO INTERNO: No encontrei \"%s\" na tabela de smbolos", sp->s_name);

		}	/* end for */

	}	/* end if (h.h_ssize) */

	/*
	 *	Verifica se h relocao para o TEXT
	 */
	rt_offset = sizeof (HEADER) + h.h_tsize + h.h_dsize + h.h_ssize;

	if (h.h_tsize != 0)
	{
		relocate
		(
			nm,
			mod_fd,
			"TEXT",
			h.h_tsize,
			sizeof (HEADER),
			h.h_rtsize,
			rt_offset,
			vector_image,
			vector_elem,
			text_fd,

			rt_fp,
			&TEXT_REL_SIZE,
			&text_offset,

			0
		);
	}

	/*
	 *	Verifica se h relocao para o DATA
	 */
	if (h.h_dsize != 0)
	{
		relocate
		(
			nm,
			mod_fd,
			"DATA",
			h.h_dsize,
			sizeof (HEADER) + h.h_tsize,
			h.h_rdsize,
			rt_offset + h.h_rtsize,
			vector_image,
			vector_elem,
			data_fd,

			rd_fp,
			&DATA_REL_SIZE,
			&data_offset,

			TEXT_SIZE
		);
	}

	text_addr += h.h_tsize;
	data_addr += h.h_dsize;
	bss_addr  += h.h_bsize;

	/*
	 *	Fecha o arquivo
	 */
	close (mod_fd);

}	/* end generate_module_sections */

/*
 ****************************************************************
 *	Reloca uma Seo					*
 ****************************************************************
 */
void
relocate
(
	const char	*mod_nm,	/* Nome do mdulo */
	int		mod_fd,		/* Descritor do mdulo */
	const char	*sec_nm,	/* Nome da seo ("TEXT/DATA") */
	unsigned	sec_sz,		/* Tamanho da seo a relocar */
	off_t		sec_off,	/* Deslocamento da seo a relocar */
	unsigned	rel_sz,		/* Tamanho da relocao dada */
	off_t		rel_off,	/* Deslocamento da seo de relocao */
	GSYM	* const vector[],	/* Endereo do vetor de ponteiros */
	int		vector_elem,	/* Nmero de elementos do vetor */
	int		fd,		/* Descritor da seo a gravar */

	FILE		*new_rel_fp,	/* Descritor do arquivo de relocao ("-r") */
	off_t		*new_rel_sz,	/* Tamanho da relocao gerada ("-r") */
	off_t		*new_rel_off,	/* Novo deslocamento ("-r") */

	off_t		a_out_off	/* Deslocamento do incio do objeto ("-h") */
)
{
	char		*sec_image;
	const char	*end_sec_image;
	void		*rel_image;
	const RELOC	*end_rel_image;

	GSYM		*gp;
	RELOC		*rp;
	long		r = 0;
	char		*bp;
	unsigned int	index;
	off_t		offset;
	char		write_rel = (rflag || bflag);

	/*
	 *	Aloca memria e le a seo indicada
	 */
	sec_image = alloca (sec_sz);

	lseek (mod_fd, sec_off, SEEK_SET);

	if (read (mod_fd, sec_image, sec_sz) != sec_sz)
		{ error ("*@Erro na leitura do %s do mdulo \"%s\"", sec_nm, mod_nm); return; }

	end_sec_image = sec_image + sec_sz;

	/*
	 *	Verifica se h relocao
	 */
	if (rel_sz != 0)
	{
		/*
		 *	Aloca memria e le a relocao
		 */
		rel_image = alloca (rel_sz);

		lseek (mod_fd, rel_off, SEEK_SET);

		if (read (mod_fd, rel_image, rel_sz) != rel_sz)
			{ error ("*@Erro na leitura de R%c do mdulo \"%s\"", *sec_nm, mod_nm); return; }

		end_rel_image = rel_image + rel_sz;

		/*
		 *	Realiza a Relocao
		 */
		for (rp = rel_image; rp < end_rel_image; rp++)
		{
			int		rel_len = rp->r_flags & RSZMASK;
			int		rel_seg = rp->r_flags & RSEGMASK;
			int		rel_code = 0;

			if ((offset = rp->r_pos) >= sec_sz)
			{
				error
				(	"@Posio %d de relocao invlido na "
					"seo %s do mdulo \"%s\"",
					offset, sec_nm, mod_nm
				);
				continue;
			}

			bp = sec_image + offset; gp = NOGSYM;

			/*
			 ****** Obtem o valor do elemento a ser relocado ****************
			 */
			switch (rel_len)
			{
			/*** case RBYTE: ***/
			   /***	r = *bp; ***/
			   /***	break; ***/

			    case RWORD:
				r = *((short *)bp);
				break;

			    case RLONG:
				r = *((long *)bp);
				break;

			    default:
				error
				(	"@Tamanho %04X de relocao invlido na "
					"seo %s do mdulo \"%s\"",
					rel_len, sec_nm, mod_nm
				);

			}	/* end switch (rel_len) */

			/*
			 ****** Obtem o novo valor do elemento **************************
			 */
			switch (rel_seg)
			{
			    case RTEXT:
				r += text_rel;
				break;

			    case RDATA:
				r += data_rel;
				break;

			    case RBSS:
				r += bss_rel;
				break;

			    case REXTREL:	/* "call" relativo ==> Muito sutil */
				r -= text_rel;
			    case REXT:
				if
				(	(index = rp->r_symbol) >= vector_elem ||
					(gp = vector[index]) == NOGSYM
				)
				{
					error
					(	"@ndice %d de relocao invlido na "
						"seo %s do mdulo \"%s\"",
						index, sec_nm, mod_nm
					);
					continue;
				}

				/* Verifica o que fazer se o smbolo ainda no foi definido */

				if (IS_UNDEF_OR_COMMON (&gp->s))
				{
					rel_code = 1;	/* Smbolo continua indefinido */

					if (shared_lib_given)
					{
						if (rel_len != RLONG) error
						(
							"@Smbolo \"%s\" com relocao NO longa na "
							"seo %s do mdulo \"%s\"",
							gp->s.s_name, sec_nm, mod_nm
						);

						insert_external_reference
						(	gp,
							a_out_off + *new_rel_off + offset
						);
					}

					break;
				}

				/* O Smbolo j foi definido */

				rel_code = 2;	/* Smbolo definido */

				r += gp->s.s_value;

				break;

			    default:
				error
				(	"@Seo %04X invlida de relocao na "
					"seo %s do mdulo \"%s\"",
					rel_seg, sec_nm, mod_nm
				);

			}	/* end switch (rel_seg) */

			/*
			 ****** Armazena o novo valor do elemento ***********************
			 */
			switch (rel_len)
			{
			/*** case RBYTE: ***/
			   /***	*bp = r; ***/
			   /***	break; ***/

			    case RWORD:
				*((short *)bp) = r;
				break;

			    case RLONG:
				*((long *)bp) = r;
				break;
			}

			/*
			 ****** Escreve a relocao (se for o caso) *********************
			 */
			if (!write_rel)
				continue;

			switch (rel_code)
			{
				/*
				 *	Relocao sem smbolo: Mantm no mesmo estilo
				 */
			    case 0:
				break;

				/*
				 *	O smbolo continua indefinido: altera o nmero do smbolo
				 */
			    case 1:
				rp->r_symbol = gp->s_index;
				break;

				/*
				 *	O smbolo foi definido
				 */
			    case 2:
				if (rel_seg == REXTREL)		/* REXTREL resolvido: No h mais relocao */
					continue;

				/* Smbolo definido: Transforma em relocao sem smbolo */

				rp->r_flags &= ~RSEGMASK;

				switch (gp->s.s_type)
				{
				    case EXTERN|ABS:		/* ABS resolvido: No h mais relocao */
					continue;

				    case EXTERN|TEXT:
					rp->r_flags |= RTEXT;
					break;

				    case EXTERN|DATA:
					rp->r_flags |= RDATA;
					break;

				    case EXTERN|BSS:
					rp->r_flags |= RBSS;
					break;

				    default:
					error
					(	"@Tipo de smbolo %04X de relocao invlido na "
						"seo %s do mdulo \"%s\"",
						gp->s.s_type, sec_nm, mod_nm
					);

				}	/* end switch (gp->s.s_type) */

				rp->r_symbol = 0;

			}	/* end switch (rel_code) */

			rp->r_pos += *new_rel_off;

			if (fwrite (rp, sizeof (RELOC), 1, new_rel_fp) != 1)
				error ("*@Erro ao gravar R%c", *sec_nm);

			*new_rel_sz += sizeof (RELOC);

		}	/* end for */

	}	/* end if (rel_sz != 0) */

	/*
	 *	Escreve & libera o espaco reservado para a seo
	 */
	if (write (fd, sec_image, sec_sz) != sec_sz)
		error ("*@Erro na escrita da seo \"%s\"", sec_nm);

	*new_rel_off += sec_sz;

}	/* end relocate */

/*
 ****************************************************************
 *	Insere a relocao na tabela de referncias externas	*
 ****************************************************************
 */
void
insert_external_reference (GSYM *gp, off_t offset)
{
	if (gp->s_ref_len >= gp->s_ref_sz)
	{
		gp->s_ref_sz += REF_VEC_INCR;

		if ((gp->s_ref_vec = realloc (gp->s_ref_vec, gp->s_ref_sz * sizeof (off_t *))) == NULL)
			error (NOSTR);
	}

	gp->s_ref_vec[gp->s_ref_len++] = offset;

	external_reference_cnt++;

}	/* end insert_external_reference */
