/*
 ****************************************************************
 *								*
 *			mesg.c					*
 *								*
 *	Habilita ou desabilita a recepo de mensagens		*
 *								*
 *	Verso	1.0.0, de 30.05.86				*
 *		3.0.0, de 23.05.97				*
 *								*
 *	Mdulo: mesg						*
 *		Utilitrios Bsicos				*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Multiprocessado		*
 *		Copyright  2000 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <stdlib.h>
#include <stat.h>
#include <errno.h>

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
const char	pgversion[] =  "Verso:	3.0.0, de 23.05.97";

#define	elif	else if

entry int	exit_code = 0;	/* Cdigo de retorno */

entry int	vflag;		/* Verbose */
entry int	nflag;		/* Desabilita */
entry int	sflag;		/* Habilita */

/*
 ****** Prottipos de funes ***********************************
 */
void		help (void);

/*
 ****************************************************************
 *	Habilita ou desabilita a recepo de mensagens		*
 ****************************************************************
 */
int
main (int argc, register const char *argv[])
{
	register int	opt;
	register char	*tty_nm;
	STAT		s;

	/*
	 *	Analisa as opes
	 */
	while ((opt = getopt (argc, argv, "nNsSyYvH")) != EOF)
	{
		switch (opt)
		{
		    case 'n':			/* Desabilita */
		    case 'N':
			if (sflag)
				help ();
			nflag++;
			break;

		    case 's':			/* Habilita */
		    case 'S':
		    case 'y':
		    case 'Y':
			if (nflag)
				help ();
			sflag++;
			break;

		    case 'v':			/* Verbose */
			vflag++;
			break;

		    case 'H':			/* Help */
			help ();

		    default:			/* Erro */
			putc ('\n', stderr);
			help ();

		}	/* end switch */

	}	/* end while */

	argv += optind;
	argc -= optind;

	/*
	 *	Obtm o estado do terminal
	 */
	if ((tty_nm = ttyname (2 /* == STDERR */)) == NOSTR)
	{
		fprintf
		(	stderr,
			"%s: No consegui obter o nome do terminal (%s)\n",
			pgname, strerror (errno)
		);
		exit (2);
	}

	if (stat (tty_nm, &s) < 0)
	{
		fprintf
		(	stderr,
			"%s: No consegui obter o estado do terminal \"%s\" (%s)\n",
			pgname, tty_nm, strerror (errno)
		);
		exit (2);
	}

	/*
	 *	Imprime o estado atual
	 */
	if (nflag + sflag == 0)
	{
		if (s.st_mode & S_IWOTH)
		{
			fprintf (stderr, "Recepo habilitada\n");
		   /***	exit_code = 0; ***/
		}
		else
		{
			fprintf (stderr, "Recepo desabilitada\n");
			exit_code = 1;
		}
	}

	/*
	 *	Altera o estado
	 */
	if   (nflag)
	{
		exit_code = 1;

		if (chmod (tty_nm, s.st_mode & ~(S_IWGRP|S_IWOTH)) < 0)
		{
			fprintf
			(	stderr,
				"%s: No consegui mudar o estado do terminal \"%s\" (%s)\n",
				pgname, tty_nm, strerror (errno)
			);
			exit_code = 2;
		}
	}
	elif (sflag)
	{
	   /***	exit_code = 0; ***/

		if (chmod (tty_nm, s.st_mode|S_IWGRP|S_IWOTH) < 0)
		{
			fprintf
			(	stderr,
				"%s: No consegui mudar o estado do terminal \"%s\" (%s)\n",
				pgname, tty_nm, strerror (errno)
			);
			exit_code = 2;
		}
	}

	return (exit_code);

}	/* end mesg */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
		"%s - habilita ou desabilita a recepo de mensagens\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"\t%s [-nNsSyY]\n",
		pgname, pgversion, pgname
	);
	fprintf
	(	stderr,
		"\nOpes:"
		"\t-n (N)      : Desabilita a recepo\n"
		"\t-s (S, y, Y): Habilita a recepo\n"
		"\nObs.:\tSem opes, imprime o estado atual\n"
	);

	exit (2);

}	/* end help */
