/*
 ****************************************************************
 *								*
 *			semafree.c				*
 *								*
 *	Libera o semforo pelo qual o processo est esperando	*
 *								*
 *	Verso	2.1.0, de 03.12.88				*
 *		3.0.0, de 30.06.97				*
 *								*
 *	Mdulo: semafree					*
 *		Utilitrios Bsicos				*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2000 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/kcntl.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
const char	pgversion[] =  "Verso: 3.0.0, de 30.06.97";

#define	NOSTR	(char *)NULL
#define	NOVOID	(void *)NULL
#define	elif	else if
#define	EVER	;;
#define STR(x)	# x
#define XSTR(x)	STR (x)

entry int	vflag;		/* Verbose */

/*
 ****** Prottipos de funes ***********************************
 */
void		help (void);

/*
 ****************************************************************
 *	Libera o semforo pelo qual o processo est esperando	*
 ****************************************************************
 */
int
main (int argc, register const char *argv[])
{
	int		opt, pid;
	const char	*str;

	/*
	 *	Analisa as opes
	 */
	while ((opt = getopt (argc, argv, "vH")) != EOF)
	{
		switch (opt)
		{
		    case 'v':			/* Verbose */
			vflag++;
			break;

		    case 'H':			/* Help */
			help ();

		    default:			/* Erro */
			putc ('\n', stderr);
			help ();

		}	/* end switch */

	}	/* end while */

	argv += optind;
	argc -= optind;

	/*
	 *	Retira o nmero do processo
	 */
	if (argc != 1)
		help ();

	if ((pid = strtol (argv[0], &str, 0)) < 0 || *str != '\0')
	{
		fprintf
		(	stderr,
			"%s: Nmero de processo \"%s\" invlido\n",
			pgname, argv[0]
		);
		exit (1);
	}

	/*
	 *	Tenta liberar o recurso
	 */
	if (kcntl (K_SEMAFREE, pid) < 0)
	{
		fprintf
		(	stderr,
			"%s: No consegui liberar o recurso (%s)\n",
			pgname, strerror (errno)
		);
		exit (1);
	}

	return (0);

}	/* end semafree */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
		"%s - libera o semforo pelo qual o processo est esperando\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"\t%s [-v] <pid>\n",
		pgname, pgversion, pgname
	);
	fprintf
	(	stderr,
		"\nOpes:"
		"\t-v: verboso\n"
	);

	exit (2);

}	/* end help */
