/*
 ****************************************************************
 *								*
 *			sort.c					*
 *								*
 *	Ordena e/ou intercala arquivos				*
 *								*
 *	Verso	1.0.0, de 03.10.86				*
 *		3.0.0, de 05.07.97				*
 *								*
 *	Mdulo: sort						*
 *		Utilitrios Bsicos				*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Multiprocessado		*
 *		Copyright  2000 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/syscall.h>

#define	DEBUG
#ifdef	DEBUG
#include <sys/kcntl.h>
#endif	DEBUG

#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include "../h/sort.h"

/*
 ****************************************************************
 *	Definies globais					*
 ****************************************************************
 */
const char	pgversion[] =  "Verso:	3.0.0, de 05.07.97";

#define	BLKMEM	4096

/*
 ****** Indicadores *********************************************
 */
entry int	vflag;		/* Verbose */
entry int	mflag;		/* Intercalao */
entry int	cflag;		/* Verifica se j est ordenado */
entry int	tflag = 'a';	/* Cdigo de caracteres */
entry int	rflag;		/* Ordenao reversa */

/*
 ****** Variveis globais ***************************************
 */
entry char	delim = ' ';	/* Delimitador entre as chaves */
entry int	numkeys;	/* No. de chaves dadas */
entry FILE	*out_fp = stdout; /* Arquivo de sada */

/*
 ****************************************************************
 *	Ordena e/ou intercala arquivos				*
 ****************************************************************
 */
int
main (int argc, register const char *argv[])
{
	register int	opt;
	const char	*out_nm = NOSTR;

	/*
	 *	Analisa as opes
	 */
	while ((opt = getopt (argc, argv, "vmcraidfo:t:k:H")) != EOF)
	{
		switch (opt)
		{
		    case 'v':			/* Verbose */
			vflag++;
			break;

		    case 'm':			/* Intercalao */
			mflag++;
			break;

		    case 'c':			/* J est ordenado? */
			cflag++;
			break;

		    case 'r':			/* Ordenao reversa */
			rflag++;
			break;

		    case 'a':			/* cdigo LEXICOGRFICO */
			tflag = 'a';
			break;

		    case 'i':			/* cdigo ISO */
			tflag = 'i';
			break;

		    case 'd':			/* cdigo DICIONRIO */
			tflag = 'd';
			break;

		    case 'f':			/* cdigo FOLD */
			tflag = 'f';
			break;

		    case 'o':			/* Nome do arquivo de sada */
			out_nm = optarg;
			break;

		    case 't':			/* Delimitador */
			delim = *optarg;
			break;

		    case 'k':			/* Chaves */
			get_keys (optarg);
			break;

		    case 'H':			/* Help */
			help ();

		    default:			/* Erro */
			putc ('\n', stderr);
			help ();

		}	/* end switch */

	}	/* end while */

	/*
	 *	Pequena consistncia
	 */
	argv += optind;
	argc -= optind;

	if (!mflag && argc > 1)
		help ();

	/*
	 *	Abre o arquivo de sada
	 */
	if (out_nm != NOSTR && (out_fp = fopen (out_nm, "w")) == NOFILE)
	{
		fprintf
		(	stderr,
			"%s: No consegui criar \"%s\" (%s)\n",
			pgname, out_nm, strerror (errno)
		);
		exit (1);
	}

	/*
	 *	Escolhe a rotina de comparao e atribui em "compar"
	 */
	set_cmp_function ();

	/*
	 *	Se foi dada a opo "-m" ento trata-se de intercalao
	 */
	if (mflag)
		merge (argv);
	else
		sort (argv[0]);

#ifdef	DEBUG
	/*
	 *	Imprime o total de memria
	 */
	if (vflag)
	{
		KPSZ	kpsz;
	
		kcntl (K_GETPSZ, &kpsz);

		fprintf
		(	stderr,
			"text = %d, data+bss = %d, stack = %d\n",
			kpsz.k_tsize, kpsz.k_dsize, kpsz.k_ssize
		);
	}
#endif	DEBUG

	return (0);

}	/* end main */

/*
 ****************************************************************
 *	Ordena							*
 ****************************************************************
 */
void
sort (const char *file_nm)
{
	register int	c, file_fd;
	register char	*ep, *cp;
	register char	**tp, **te;
	char		*file_area = NOSTR;
	char		**tag_area = NOVOID;
	int		tag_sz;
	off_t		file_sz;
	STAT		s;

	/*
	 *	L o arquivo (comea a rea com um "\n")
	 */
	if (file_nm != NOSTR)
	{
		if ((file_fd = open (file_nm, O_RDONLY)) < 0)
		{
			fprintf
			(	stderr,
				"%s: No consegui abrir \"%s\" (%s)\n",
				pgname, file_nm, strerror (errno)
			);
			exit (1);
		}

		if (fstat (file_fd, &s) < 0)
		{
			fprintf
			(	stderr,
				"%s: No consegui obter o estado de \"%s\" (%s)\n",
				pgname, file_nm, strerror (errno)
			);
			exit (1);
		}

		if (!S_ISREG (s.st_mode))
		{
			fprintf (stderr, "%s: \"%s\" no  regular\n", pgname, file_nm);
			exit (1);
		}

		file_sz = s.st_size + 1;	/* No esquece o "\n" extra */

		if ((file_area = malloc (file_sz)) == NOSTR)
			{ fprintf (stderr, "Memria esgotada\n"); exit (1); }

		file_area[0] = '\n';

		if (s.st_size && read (file_fd, file_area + 1, s.st_size) != s.st_size)
		{
			fprintf
			(	stderr,
				"%s: No consegui ler \"%s\" (%s)\n",
				pgname, file_nm, strerror (errno)
			);
			exit (1);
		}
	}
	else		/* PIPE */
	{
		file_sz = 0;	ep = cp = NOSTR;

		while ((c = getchar ()) != EOF)
		{
			if (cp >= ep)
			{
				if ((file_area = realloc (file_area, file_sz + BLKMEM)) == NOSTR)
					{ fprintf (stderr, "Memria esgotada\n"); exit (1); }

				cp = file_area + file_sz;
				ep = cp + BLKMEM;

				if (file_sz == 0) /* No esquece o "\n" extra */
					*cp++ = '\n';

				file_sz += BLKMEM;
			}

			*cp++ = c;
		}

		file_sz = cp - file_area;
	}
	
	if (file_sz <= 1)
	{
		fprintf (stderr, "%s: Arquivo de tamanho NULO?\n", pgname);
		exit (1);
	}

	/*
	 *	Cria os "tags"
	 */
	tag_sz = 0;	tp = te = NOVOID;

	for (cp = file_area, ep = cp + file_sz; cp < ep; cp++)
	{
		if (*cp == '\n')
		{
			if (tp >= te)
			{
				if ((tag_area = realloc (tag_area, tag_sz * sizeof (char *) + BLKMEM)) == NOVOID)
					{ fprintf (stderr, "Memria esgotada\n"); exit (1); }
	
				tp = tag_area + tag_sz;
				te = tp + (BLKMEM / sizeof (char *));
	
				tag_sz += (BLKMEM / sizeof (char *));
			}

			*cp = '\0';		*tp++ = cp + 1;
		}
	}

	tp--; 	tag_sz = tp - tag_area; 	/* O ltimo no existe */

	te = tp;	/* Final dos "tags" */

	if (vflag)
		fprintf (stderr, "No. de elementos = %d\n", tag_sz);

	/*
	 *	 apenas para verificar a ordenao?
	 */
	if (cflag) for (tp = tag_area; /* abaixo */; tp++)
	{
		if (tp >= te - 1)
		{
			fprintf
			(	stderr, file_nm ?
				"O arquivo \"%s\" j est ordenado\n" :
				"O arquivo j est ordenado\n", file_nm
			);
			exit (0);
		}

		if (compar (tp[0], tp[1]) > 0)
		{
			fprintf
			(	stderr, file_nm ?
				"O arquivo \"%s\" NO est ordenado\n" :
				"O arquivo NO est ordenado\n", file_nm
			);
			exit (1);
		}
	}

	/*
	 *	Ordena
	 */
	qtsort (tag_area, tag_sz, compar);

	/*
	 *	Escreve o arquivo ordenado
	 */
	for (tp = tag_area; tp < te; tp++)
		fnputs (tp[0], out_fp);

}	/* end sort */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
		"%s - ordena e/ou intercala arquivos\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"    %s [-vcraidf]   [-o <sada>] "
		"[-t <d>] [-k \"x:y.zot ... \"] [<entrada>]\n"
		"    %s -m [-vraidf] [-o <sada>] "
		"[-t <d>] [-k \"x:y.zot ... \"] <entrada> ...\n",
		pgname, pgversion, pgname, pgname
	);
	fprintf
	(	stderr,
		"\nOpes:"
		"\t-v: Verboso\n"
		"\t-m: Intercalao (\"merge\")\n"
		"\t-c: Verifica se o arquivo j est ordenado\n"
		"\t-r: Ordenao reversa (descendente)\n"
		"\t-a: Tipo de ordenao LEXICOGRFICO (tipo por omisso)\n"
		"\t-i: Tipo de ordenao ISO\n"
		"\t-d: Tipo de ordenao DICIONRIO\n"
		"\t-f: Tipo de ordenao FOLD (mausculas = minsculas)\n"
	);
	fprintf
	(	stderr,
		"\t-o: Escreve o arquivo ordenado em <sada> ao invs de \"stdout\"\n"
		"\t-t  Utiliza o caracter <d> ao invs do branco como delimitador de campos\n"
		"\t-k: Especifica chaves:\n"
		"\t\t x: No. do campo\n"
		"\t\t y: Incio da chave dentro do campo\n"
		"\t\t z: Largura da chave\n"
		"\t\t o: Ordem ascendente/descendente (\"a\" ou \"d\")\n"
		"\t\t t: Tipo (\"a\", \"i\", \"d\", ou \"f\"), como acima\n"
		"\nObs.:\tSem opes: ordem ascendente, tipo LEXICOGRFICO, sada em \"stdout\"\n"
	);

	exit (2);

}	/* end help */
