/*
 ****************************************************************
 *								*
 *			uname.c					*
 *								*
 *	Imprime a identificao do sistema			*
 *								*
 *	Verso	1.0.0, de 09.02.86				*
 *		3.0.0, de 13.07.97				*
 *								*
 *	Mdulo: uname						*
 *		Utilitrios Bsicos				*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Multiprocessado		*
 *		Copyright  2000 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/syscall.h>
#include <sys/utsname.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
const char	pgversion[] =  "Verso:	3.0.0, de 13.07.97";

#define	NOVOID	(void *)NULL
#define	elif	else if
#define	EVER	;;
#define STR(x)	# x
#define XSTR(x)	STR (x)

entry int	vflag;		/* Verbose */

/*
 ****** Prottipos de funes ***********************************
 */
void		help (void);

/*
 ****************************************************************
 *	Imprime a identificao do sistema			*
 ****************************************************************
 */
int
main (int argc, register const char *argv[])
{
	register int	opt;
	UTSNAME		uts;

	/*
	 *	Analisa as opes
	 */
	while ((opt = getopt (argc, argv, "vH")) != EOF)
	{
		switch (opt)
		{
		    case 'v':			/* Verbose */
			vflag++;
			break;

		    case 'H':			/* Help */
			help ();

		    default:			/* Erro */
			putc ('\n', stderr);
			help ();

		}	/* end switch */

	}	/* end while */

	argv += optind;
	argc -= optind;

	/*
	 *	Pequena consistncia
	 */
	if (argc > 0)
		help ();

	/*
	 *	Obtm a identificao do sistema
	 */
	if (uname (&uts) < 0)
	{
		fprintf
		(	stderr,
			"%s: No consegui obter a identificao do sistema (%s)\n",
			pgname, strerror (errno)
		);
		exit (1);
	}

	/*
	 *	Imprime a identificao do sistema (prosa)
	 */
	if (!vflag)
	{
		printf
		(	"%s %s - Verso %s de %s s %s em um %s (%s)\n",
			uts.uts_sysname, 	uts.uts_systype,
			uts.uts_version, 	uts.uts_date,
			uts.uts_time, 		uts.uts_machine,
			uts.uts_nodename
		);

#ifdef	ICA
		printf
		(	"\nNo. de srie: %s,   Uso autorizado para: %s, %s\n",
			uts.uts_sysserno, 	uts.uts_customer,
			uts.uts_depto
		);
#endif	ICA

		exit (0);
	}

	/*
	 *	Imprime a identificao do sistema (estrutura)
	 */
	printf
	(	"sysname 	= %s\n"	/* Nome do Sistema */
		"systype 	= %s\n"	/* Tipo do Sistema */
		"nodename	= %s\n"	/* Nome do n */
		"version 	= %s\n"	/* Verso */
		"date		= %s\n"	/* Data */
		"time		= %s\n"	/* Tempo */
		"machine 	= %s\n"	/* Computador */
		"customer	= %s\n"	/* Nome do Cliente */
		"depto		= %s\n"	/* Nome do Departamento */
		"sysserno	= %s\n",/* Numero de serie */

		uts.uts_sysname,	/* Nome do Sistema */
		uts.uts_systype,	/* Tipo do Sistema */
		uts.uts_nodename,	/* Nome do n */
		uts.uts_version,	/* Verso */
		uts.uts_date,		/* Data */
		uts.uts_time,		/* Tempo */
		uts.uts_machine,	/* Computador */
		uts.uts_customer,	/* Nome do Cliente */
		uts.uts_depto,		/* Nome do Departamento */
		uts.uts_sysserno	/* Numero de serie do Sistema */
	);

	return (0);

}	/* end uname */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
	 	"%s - imprime a identificao do sistema\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"\t%s [-v]\n",
		pgname, pgversion, pgname
	);

	fprintf
	(	stderr,
		"\nOpes:"
		"\t-v: Formato da estrutura \"uts\"\n"
	);

	exit (2);

}	/* end help */
