/*
 ****************************************************************
 *								*
 *			walk.c					*
 *								*
 * 	Caminha em rvores do sistema de arquivos		*
 *								*
 *	Verso	1.0.0, de 28.09.86				*
 *		4.2.0, de 25.04.02				*
 *								*
 *	Mdulo: walk						*
 *		Utilitrios bsicos				*
 *		Categoria B					*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2002 NCE/UFRJ - tecle "man licena"	*
 * 								*
 ****************************************************************
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ftw.h>
#include <fcntl.h>
#include <errno.h>

/*
 ****************************************************************
 *	Variveis e Definies globais				*
 ****************************************************************
 */
const char	pgversion[] =  "Verso:	4.2.0, de 25.04.02";

#define	EVER	;;
#define	elif	else if
#define STR(x)	# x
#define XSTR(x)	STR (x)

#define	NPAT	20	/* No. maximo de Padres */

/*
 ******	Variveis globais ***************************************
 */
entry int	rflag;		/* Arquivos regulares */
entry int	dflag;		/* Diretrios  */
entry int	bflag;		/* Arquivos especiais de blocos */
entry int	cflag;		/* Arquivos especiais de caracteres */
entry int	lflag;		/* Elos simblicos */
entry int	fflag;		/* FIFOs  */
entry int	mflag;		/* Prefixo */
entry int	Lflag;		/* Evite LINKS flag */
entry int	dotflag;	/* No ignora ".old" ... */
entry int	Nflag;		/* Nomes dos pgms no <stdin> */
entry int	vflag;		/* Verbose */

const char	**color_vector;	/* Para as cores dos modos */

/*
 ******	Os padres **********************************************
 */
entry const char *inc_pat[NPAT+1];	/* Os Padres de incluso */
entry int	inc_pati;

entry const char *exc_pat[NPAT+1];	/* Os Padres de excluso */
entry int	exc_pati;

/*
 ******	Prottipos de funes ***********************************
 */
int		accept (const char *, STAT *, int);
void		put_s_option (void);
int		put_p_option (char *, const char *[], int *);
int		pattern_accept (const char *);
void		process (const char *, STAT *);
int		proclink (STAT *);
void		help (void);

/*
 ****************************************************************
 *	Estruturas						*
 ****************************************************************
 */
typedef struct litem	LITEM;

struct litem
{
	dev_t	l_dev;		/* Dispositivo */
	ino_t	l_ino;		/* No. do Inode */
	LITEM	*l_next;	/* o Proximo */
};

entry LITEM	*linklist;	/* Comeco da Lista */

/*
 ****************************************************************
 * 	Caminha em rvores do sistema de arquivos		*
 ****************************************************************
 */
int
main (int argc, const char *argv[])
{
	int		opt;

	/*
	 *	Analisa as opes
	 */
	while ((opt = getopt (argc, argv, "rdbclfamL.Nsp:P:vH")) != EOF)
	{
		switch (opt)
		{
		    case 'r':		/* Arquivos regulares */
			rflag++;
			continue;

		    case 'd':		/* Diretrios */
			dflag++;
			continue;

		    case 'b':		/* Especial de blocos */
			bflag++;
			continue;

		    case 'c':		/* Especial de caracteres */
			cflag++;
			continue;

		    case 'l':		/* Elos Simblicos */
			lflag++;
			continue;

		    case 'f':		/* FIFOs */
			fflag++;
			continue;

		    case 'a':		/* Todos tipos */
			rflag++;
			dflag++;
			bflag++;
			cflag++;
			fflag++;
			continue;

		    case 'm':			/* Prefixo */
			mflag++;
			break;

		    case 'L':			/* evite LINKS */
			Lflag++;
			break;

		    case '.':			/* No ignora ".old" ... */
			dotflag++;
			break;

		    case 'N':			/* Nomes do <stdin> */
			Nflag++;
			break;

		    case 's':			/* Padres = 	*.[cshryv] e */
			put_s_option ();
			break;

		    case 'p':		/* Padres de incluso */
			put_p_option ((char *)optarg, inc_pat, &inc_pati);
			break;

		    case 'P':		/* Padres de excluso */
			put_p_option ((char *)optarg, exc_pat, &exc_pati);
			break;

		    case 'v':		/* verbose flag */
			vflag++;
			continue;

		    case 'H':			/* Help */
			help ();

		    default:			/* Erro */
			fputc ('\n', stderr);
			help ();

		}	/* end switch */

	}	/* end while */ 

	/*
	 *	Ajustes
	 */
	argv += optind;

	if (rflag + dflag + bflag + cflag + lflag + fflag == 0)
		{ rflag++; bflag++; cflag++; lflag++; fflag++; }

	/*
	 *	Processa "modecolor"
	 */
	modecolor (isatty (fileno (stdout)), (void *)NULL, &color_vector);

	/*
	 *	Analisa o Nflag
	 */
	if (Nflag && *argv != NOSTR)
		{ error ("Os argumentos suprfluos sero ignorados"); *argv = NOSTR; }

	/*
	 *	Procura os argumentos
	 */
	if (*argv == NOSTR) 		/* No foram dados argumentos: l do "stdin" */
	{
		char		*area = alloca (1024);

		while (fngets (area, 1024, stdin) != NOSTR)
			ftw (area, accept);
	}
	else 				/* Foram dados argumentos */
	{
		for (/* vazio */; *argv; argv++)
			ftw (*argv, accept);
	}

	return (0);

}	/* end walk */

/*
 ****************************************************************
 *	Processa um nome 					*
 ****************************************************************
 */
int
accept (const char *file_nm, STAT *sp, int status)
{
	if   (status == FTW_NS)
	{
		error ("*No consegui obter o estado de \"%s\"", file_nm);
		return (0);
	}
	elif (status == FTW_DNR)
	{
		error ("*No consegui percorrer \"%s\"", file_nm);
	}

	/*
	 *	Seleciona os tipos de arquivos
	 */
	switch (sp->st_mode & S_IFMT)
	{
	    case S_IFREG:
		if (!rflag)
			return (0);
		break;

	    case S_IFDIR:	/* Diretrios ".old": pula toda subrvore */
		if (!dotflag)
		{
			if (patmatch (file_nm, ".!.*") || patmatch (file_nm, "*/.!.*"))
				return (1);
		}

		if (!dflag)
			return (0);
		break;

	    case S_IFIFO:
		if (!fflag)
			return (0);
		break;

	    case S_IFBLK:
		if (!bflag)
			return (0);
		break;

	    case S_IFCHR:
		if (!cflag)
			return (0);
		break;

	    case S_IFLNK:
		if (!lflag)
			return (0);
		break;

	    default:
		error ("Tipo invlido de arquivo: \"%s\"", file_nm);
		return (0);

	}	/* end switch */


	/*
	 *	Analisa os padres
	 */
	if (pattern_accept (file_nm))
		process (file_nm, sp);

	return (0);

}	/* end accept */

/*
 ****************************************************************
 *	Executa a opo "-s" (padres de "C")			*
 ****************************************************************
 */
void
put_s_option (void)
{
	put_p_option ("*.[cshryv]", inc_pat, &inc_pati);
	put_p_option ("*[Mm]akefile", inc_pat, &inc_pati);

}	/* end put_s_option */

/*
 ****************************************************************
 *	Executa a opo "-p" (outros padres)			*
 ****************************************************************
 */
int
put_p_option (char *cp, const char *pat_vec[], int *pati_ptr)
{
	char		c;

	for (EVER)
	{
		while ((c = *cp) == ' ' || c == '\t' )
			cp++;

		if (c == '\0')
			return (0);

		if (*pati_ptr >= NPAT)
		{
			printf
			(	"%s: No momento somente aceitamos "
				XSTR (NPAT) " padres\n\n",
				pgname
			);
			help ();
		}

		pat_vec[(*pati_ptr)++] = cp++;

		while ((c = *cp) != '\0' && c != ' ' && c != '\t')
			cp++;

		if (c == '\0')
			return (0);

		*cp++ = '\0';

	}	/* end for (EVER) */

}	/* end put_p_option */

/*
 ****************************************************************
 *	Verifica se o nome do arquivo satisfaz um dos padres	*
 ****************************************************************
 */
int
pattern_accept (const char *file_nm)
{
	const char	**pp;

	/*
	 *	Testa em primeiro lugar os pontos
	 */
	if (!dotflag)
	{
		if (patmatch (file_nm, ".!.*") || patmatch (file_nm, "*/.!.*"))
			return (0);
	}

	/*
	 *	Agora testa os padres de incluso
	 */
	if ((pp = inc_pat)[0] != NOSTR)
	{
		for (/* vazio */; /* vazio */; pp++)
		{
			if (*pp == NOSTR)
				return (0);

			if (patmatch (file_nm, *pp))
				break;
		}
	}

	/*
	 *	Finalmente testa os padres de excluso
	 */
	if ((pp = exc_pat)[0] != NOSTR)
	{
		for (/* vazio */; *pp != NOSTR; pp++)
		{
			if (patmatch (file_nm, *pp))
				return (0);
		}
	}

	return (1);

}	/* end pattern_accept */

/*
 ****************************************************************
 *	Processa O arquivo					*
 ****************************************************************
 */
void
process (const char *file_nm, STAT *sp)
{

	if (Lflag)
	{
		if (proclink (sp) > 0)
			return;
	}

	if (mflag)
	{
		printf ("%c  ", *(modetostr (sp->st_mode)));

		if (S_ISREG (sp->st_mode))
			printf ("%7d  ", sp->st_size);
		else
			printf ("         ");
	}

	printf
	(	"%s%s%s\n",
		color_vector[(sp->st_mode & S_IFMT) >> 12],
		file_nm,
		color_vector[MC_DEF]
	);

}	/* end process */

/*
 ****************************************************************
 *	Processa LINKs						*
 ****************************************************************
 */
int
proclink (STAT *sp)
{
	LITEM		*lp;

	if (sp->st_nlink == 1)
		return (0);

	for (lp = linklist; lp != (LITEM *)NULL; lp = lp->l_next)
	{
		if (lp->l_dev == sp->st_dev && lp->l_ino == sp->st_ino)
			return (1);
	}

	if ((lp = malloc (sizeof (LITEM))) == (LITEM *)NULL)
	{
		error ("No obtive memria para item de elo fsico");
	}
	else
	{
		lp->l_dev = sp->st_dev;
		lp->l_ino = sp->st_ino;

		lp->l_next = linklist;
		linklist = lp;
	}

	return (0);

}	/* end proclink */

/*
 ****************************************************************
 *	Resumo de utilizao do programa			*
 ****************************************************************
 */
void
help (void)
{
	fprintf
	(	stderr,
	 	"%s - caminha em rvores do sistema de arquivos\n"
		"\n%s\n"
		"\nSintaxe:\n"
		"\t%s [rdbclfamL.N] [<rvore> ...]\n\n"
		"\tValores de <PAD>: [-s] [-p \"<p>\"] [-P \"<P>\"]\n",
		pgname, pgversion, pgname
	);
	fprintf
	(	stderr,
		"\nOpes:"
		"\t--: Sem nenhuma das opes \"-rdbclfa\" lista todos os tipos "
			"\n\t    de arquivos exceto diretrios\n"
		"\t-r: Lista arquivos regulares\n"
		"\t-d: Lista diretrios\n"
		"\t-b: Lista arquivos especiais de bloco (estruturados)\n"
		"\t-c: Lista arquivos especiais de caracteres (no estruturados)\n"
		"\t-l: Lista elos simblicos\n"
		"\t-f: Lista FIFOs\n"
		"\t-a: Lista todos os tipos de arquivos (equivalente a \"-rdbcf\")\n"
		"\t-m: Imprime um prefixo com o tipo do arquivo e seu tamanho\n"
	);
	fprintf
	(	stderr,
		"\t-L: No repete arquivos com mais de um elo fsico\n"
		"\t-.: Lista tambm arquivos ocultos (\".old\" ...)\n"
		"\t-N: L os nomes das <rvore>s de \"stdin\"\n"
		"\t    Esta opo  implcita se no forem dados <rvore>s\n"
		"\t-s: Adiciona os padres dos fontes em \"C\"\n"
		"\t-p: Seleciona arquivos atravs dos padres <p> de incluso\n"
		"\t-P: Seleciona arquivos atravs dos padres <P> de excluso\n"
	);
	exit (2);

}	/* end help */
